# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause

import abc
import os
from contextlib import contextmanager
from enum import IntEnum

import llvmlite
from llvmlite import ir
from numba.cuda import types
from numba.cuda.core import config
from numba.cuda import cgutils
from numba.cuda.datamodel.models import ComplexModel, UnionModel, UniTupleModel
from numba.cuda.types.ext_types import GridGroup
from cuda.bindings import runtime


# Check if CUDA Toolkit and llvmlite support polymorphic debug info
def _get_llvmlite_version():
    """Get llvmlite version as tuple (major, minor)."""
    try:
        version_str = llvmlite.__version__
        # Parse version string like "0.46.0" or "0.46.0dev"
        parts = version_str.split(".")
        if len(parts) < 2:
            return (0, 0)
        major = int(parts[0])
        minor = int(parts[1])
        return (major, minor)
    except (IndexError, AttributeError, ValueError):
        return (0, 0)


def _check_polymorphic_debug_info_support():
    """Check if CTK and llvmlite support polymorphic debug info.

    Returns:
        tuple: (supported: bool, use_typed_const: bool)
        - supported: Whether feature is supported at all
        - use_typed_const: True for typed constant,
                           False for node reference
    """
    # runtime.getLocalRuntimeVersion() returns (cudaError_t, version_int)
    # Example: 13010 = CTK 13.1, 13020 = CTK 13.2
    _, ctk_version_number = runtime.getLocalRuntimeVersion()
    ctk_major = ctk_version_number // 1000
    ctk_minor = (ctk_version_number % 1000) // 10
    ctk_version = (ctk_major, ctk_minor)

    llvmlite_version = _get_llvmlite_version()

    # Support not available with CTK 13.1 or older
    if ctk_version <= (13, 1):
        return (False, False)

    # llvmlite > 0.45: use typed constant
    # llvmlite <= 0.45: use node reference
    use_typed_const = llvmlite_version > (0, 45)
    return (True, use_typed_const)


# Check support and determine mode
(DEBUG_POLY_SUPPORTED, DEBUG_POLY_USE_TYPED_CONST) = (
    _check_polymorphic_debug_info_support()
)

# Set config based on polymorphic debug info support
if not hasattr(config, "CUDA_DEBUG_POLY"):
    config.CUDA_DEBUG_POLY = DEBUG_POLY_SUPPORTED
if not hasattr(config, "CUDA_DEBUG_POLY_USE_TYPED_CONST"):
    config.CUDA_DEBUG_POLY_USE_TYPED_CONST = DEBUG_POLY_USE_TYPED_CONST


class DwarfAddressClass(IntEnum):
    GENERIC = 0x00
    GLOBAL = 0x01
    REGISTER = 0x02
    CONSTANT = 0x05
    LOCAL = 0x06
    PARAMETER = 0x07
    SHARED = 0x08


@contextmanager
def suspend_emission(builder):
    """Suspends the emission of debug_metadata for the duration of the context
    managed block."""
    ref = builder.debug_metadata
    builder.debug_metadata = None
    try:
        yield
    finally:
        builder.debug_metadata = ref


class AbstractDIBuilder(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def mark_variable(
        self,
        builder,
        allocavalue,
        name,
        lltype,
        size,
        line,
        datamodel=None,
        argidx=None,
    ):
        """Emit debug info for the variable."""
        pass

    @abc.abstractmethod
    def mark_location(self, builder, line):
        """Emit source location information to the given IRBuilder."""
        pass

    @abc.abstractmethod
    def mark_subprogram(self, function, qualname, argnames, argtypes, line):
        """Emit source location information for the given function."""
        pass

    @abc.abstractmethod
    def initialize(self):
        """Initialize the debug info. An opportunity for the debuginfo to
        prepare any necessary data structures.
        """

    @abc.abstractmethod
    def finalize(self):
        """Finalize the debuginfo by emitting all necessary metadata."""
        pass


class DummyDIBuilder(AbstractDIBuilder):
    def __init__(self, module, filepath, cgctx, directives_only):
        pass

    def mark_variable(
        self,
        builder,
        allocavalue,
        name,
        lltype,
        size,
        line,
        datamodel=None,
        argidx=None,
    ):
        pass

    def mark_location(self, builder, line):
        pass

    def mark_subprogram(self, function, qualname, argnames, argtypes, line):
        pass

    def initialize(self):
        pass

    def finalize(self):
        pass


_BYTE_SIZE = 8


class DIBuilder(AbstractDIBuilder):
    DWARF_VERSION = 4
    DEBUG_INFO_VERSION = 3
    DBG_CU_NAME = "llvm.dbg.cu"
    _DEBUG = False

    def __init__(self, module, filepath, cgctx, directives_only):
        self.module = module
        self.filepath = os.path.abspath(filepath)
        self.difile = self._di_file()
        self.subprograms = []
        self.cgctx = cgctx

        if directives_only:
            self.emission_kind = "DebugDirectivesOnly"
        else:
            self.emission_kind = "FullDebug"

        self.initialize()

    def initialize(self):
        # Create the compile unit now because it is referenced when
        # constructing subprograms
        self.dicompileunit = self._di_compile_unit()

    def get_dwarf_address_class(self, addrspace):
        # Map NVVM address space to DWARF address class.
        from numba.cuda.cudadrv import nvvm

        addrspace_to_addrclass_dict = {
            nvvm.ADDRSPACE_GENERIC: None,
            nvvm.ADDRSPACE_GLOBAL: DwarfAddressClass.GLOBAL,
            nvvm.ADDRSPACE_SHARED: DwarfAddressClass.SHARED,
            nvvm.ADDRSPACE_CONSTANT: DwarfAddressClass.CONSTANT,
            nvvm.ADDRSPACE_LOCAL: DwarfAddressClass.LOCAL,
        }
        return addrspace_to_addrclass_dict.get(addrspace)

    def _var_type(self, lltype, size, datamodel=None):
        if self._DEBUG:
            print(
                "-->",
                lltype,
                size,
                datamodel,
                getattr(datamodel, "fe_type", "NO FE TYPE"),
            )
        m = self.module
        bitsize = _BYTE_SIZE * size

        int_type = (ir.IntType,)
        real_type = ir.FloatType, ir.DoubleType
        # For simple numeric types, choose the closest encoding.
        # We treat all integers as unsigned when there's no known datamodel.
        if isinstance(lltype, int_type + real_type):
            if datamodel is None:
                # This is probably something like an `i8*` member of a struct
                name = str(lltype)
                if isinstance(lltype, int_type):
                    ditok = "DW_ATE_unsigned"
                else:
                    ditok = "DW_ATE_float"
            else:
                # This is probably a known int/float scalar type
                name = str(datamodel.fe_type)
                if isinstance(datamodel.fe_type, types.Integer):
                    if datamodel.fe_type.signed:
                        ditok = "DW_ATE_signed"
                    else:
                        ditok = "DW_ATE_unsigned"
                else:
                    ditok = "DW_ATE_float"
            mdtype = m.add_debug_info(
                "DIBasicType",
                {
                    "name": name,
                    "size": bitsize,
                    "encoding": ir.DIToken(ditok),
                },
            )
        elif isinstance(datamodel, ComplexModel):
            # TODO: Is there a better way of determining "this is a complex
            # number"?
            #
            # NOTE: Commented below is the way to generate the metadata for a
            # C99 complex type that's directly supported by DWARF. Numba however
            # generates a struct with real/imag cf. CPython to give a more
            # pythonic feel to inspection.
            #
            # mdtype = m.add_debug_info('DIBasicType', {
            #  'name': f"{datamodel.fe_type} ({str(lltype)})",
            #  'size': bitsize,
            # 'encoding': ir.DIToken('DW_ATE_complex_float'),
            # })
            meta = []
            offset = 0
            for ix, name in enumerate(("real", "imag")):
                component = lltype.elements[ix]
                component_size = self.cgctx.get_abi_sizeof(component)
                component_basetype = m.add_debug_info(
                    "DIBasicType",
                    {
                        "name": str(component),
                        "size": _BYTE_SIZE * component_size,  # bits
                        "encoding": ir.DIToken("DW_ATE_float"),
                    },
                )
                derived_type = m.add_debug_info(
                    "DIDerivedType",
                    {
                        "tag": ir.DIToken("DW_TAG_member"),
                        "name": name,
                        "baseType": component_basetype,
                        "size": _BYTE_SIZE
                        * component_size,  # DW_TAG_member size is in bits
                        "offset": offset,
                    },
                )
                meta.append(derived_type)
                offset += _BYTE_SIZE * component_size  # offset is in bits
            mdtype = m.add_debug_info(
                "DICompositeType",
                {
                    "tag": ir.DIToken("DW_TAG_structure_type"),
                    "name": f"{datamodel.fe_type} ({str(lltype)})",
                    "identifier": str(lltype),
                    "elements": m.add_metadata(meta),
                    "size": offset,
                },
                is_distinct=True,
            )
        elif isinstance(datamodel, UniTupleModel):
            element = lltype.element
            el_size = self.cgctx.get_abi_sizeof(element)
            element_model = next(iter(datamodel.inner_models()), None)
            basetype = self._var_type(element, el_size, datamodel=element_model)
            name = f"{datamodel.fe_type} ({str(lltype)})"
            count = size // el_size
            mdrange = m.add_debug_info(
                "DISubrange",
                {
                    "count": count,
                },
            )
            mdtype = m.add_debug_info(
                "DICompositeType",
                {
                    "tag": ir.DIToken("DW_TAG_array_type"),
                    "baseType": basetype,
                    "name": name,
                    "size": bitsize,
                    "identifier": str(lltype),
                    "elements": m.add_metadata([mdrange]),
                },
            )
        elif isinstance(lltype, ir.PointerType):
            model = getattr(datamodel, "_pointee_model", None)
            basetype = self._var_type(
                lltype.pointee, self.cgctx.get_abi_sizeof(lltype.pointee), model
            )
            mdtype = m.add_debug_info(
                "DIDerivedType",
                {
                    "tag": ir.DIToken("DW_TAG_pointer_type"),
                    "baseType": basetype,
                    "size": _BYTE_SIZE * self.cgctx.get_abi_sizeof(lltype),
                },
            )
        elif isinstance(lltype, ir.LiteralStructType):
            # Struct type
            meta = []
            offset = 0
            if datamodel is None or not datamodel.inner_models():
                name = f"Anonymous struct ({str(lltype)})"
                for field_id, element in enumerate(lltype.elements):
                    size = self.cgctx.get_abi_sizeof(element)
                    basetype = self._var_type(element, size)
                    derived_type = m.add_debug_info(
                        "DIDerivedType",
                        {
                            "tag": ir.DIToken("DW_TAG_member"),
                            "name": f"<field {field_id}>",
                            "baseType": basetype,
                            "size": _BYTE_SIZE
                            * size,  # DW_TAG_member size is in bits
                            "offset": offset,
                        },
                    )
                    meta.append(derived_type)
                    offset += _BYTE_SIZE * size  # offset is in bits
            else:
                name = f"{datamodel.fe_type} ({str(lltype)})"
                for element, field, model in zip(
                    lltype.elements, datamodel._fields, datamodel.inner_models()
                ):
                    size = self.cgctx.get_abi_sizeof(element)
                    basetype = self._var_type(element, size, datamodel=model)
                    derived_type = m.add_debug_info(
                        "DIDerivedType",
                        {
                            "tag": ir.DIToken("DW_TAG_member"),
                            "name": field,
                            "baseType": basetype,
                            "size": _BYTE_SIZE
                            * size,  # DW_TAG_member size is in bits
                            "offset": offset,
                        },
                    )
                    meta.append(derived_type)
                    offset += _BYTE_SIZE * size  # offset is in bits

            mdtype = m.add_debug_info(
                "DICompositeType",
                {
                    "tag": ir.DIToken("DW_TAG_structure_type"),
                    "name": name,
                    "identifier": str(lltype),
                    "elements": m.add_metadata(meta),
                    "size": offset,
                },
                is_distinct=True,
            )
        elif isinstance(lltype, ir.ArrayType):
            element = lltype.element
            el_size = self.cgctx.get_abi_sizeof(element)
            basetype = self._var_type(element, el_size)
            count = size // el_size
            mdrange = m.add_debug_info(
                "DISubrange",
                {
                    "count": count,
                },
            )
            mdtype = m.add_debug_info(
                "DICompositeType",
                {
                    "tag": ir.DIToken("DW_TAG_array_type"),
                    "baseType": basetype,
                    "name": str(lltype),
                    "size": bitsize,
                    "identifier": str(lltype),
                    "elements": m.add_metadata([mdrange]),
                },
            )
        else:
            # For all other types, describe it as sequence of bytes
            count = size
            mdrange = m.add_debug_info(
                "DISubrange",
                {
                    "count": count,
                },
            )
            mdbase = m.add_debug_info(
                "DIBasicType",
                {
                    "name": "byte",
                    "size": _BYTE_SIZE,
                    "encoding": ir.DIToken("DW_ATE_unsigned_char"),
                },
            )
            mdtype = m.add_debug_info(
                "DICompositeType",
                {
                    "tag": ir.DIToken("DW_TAG_array_type"),
                    "baseType": mdbase,
                    "name": str(lltype),
                    "size": bitsize,
                    "identifier": str(lltype),
                    "elements": m.add_metadata([mdrange]),
                },
            )

        return mdtype

    def mark_variable(
        self,
        builder,
        allocavalue,
        name,
        lltype,
        size,
        line,
        datamodel=None,
        argidx=None,
    ):
        arg_index = 0 if argidx is None else argidx
        m = self.module
        fnty = ir.FunctionType(ir.VoidType(), [ir.MetaDataType()] * 3)
        decl = cgutils.get_or_insert_function(m, fnty, "llvm.dbg.declare")

        mdtype = self._var_type(lltype, size, datamodel=datamodel)
        name = name.replace(".", "$")  # for gdb to work correctly
        mdlocalvar = m.add_debug_info(
            "DILocalVariable",
            {
                "name": name,
                "arg": arg_index,
                "scope": self.subprograms[-1],
                "file": self.difile,
                "line": line,
                "type": mdtype,
            },
        )
        mdexpr = m.add_debug_info("DIExpression", {})

        return builder.call(decl, [allocavalue, mdlocalvar, mdexpr])

    def mark_location(self, builder, line):
        builder.debug_metadata = self._add_location(line)

    def mark_subprogram(self, function, qualname, argnames, argtypes, line):
        name = qualname
        argmap = dict(zip(argnames, argtypes))
        di_subp = self._add_subprogram(
            name=name,
            linkagename=function.name,
            line=line,
            function=function,
            argmap=argmap,
        )
        function.set_metadata("dbg", di_subp)

    def finalize(self):
        dbgcu = cgutils.get_or_insert_named_metadata(
            self.module, self.DBG_CU_NAME
        )
        dbgcu.add(self.dicompileunit)
        self._set_module_flags()

    #
    # Internal APIs
    #

    def _set_module_flags(self):
        """Set the module flags metadata"""
        module = self.module
        mflags = cgutils.get_or_insert_named_metadata(
            module, "llvm.module.flags"
        )
        # Set *require* behavior to warning
        # See http://llvm.org/docs/LangRef.html#module-flags-metadata
        require_warning_behavior = self._const_int(2)
        if self.DWARF_VERSION is not None:
            dwarf_version = module.add_metadata(
                [
                    require_warning_behavior,
                    "Dwarf Version",
                    self._const_int(self.DWARF_VERSION),
                ]
            )
            if dwarf_version not in mflags.operands:
                mflags.add(dwarf_version)
        debuginfo_version = module.add_metadata(
            [
                require_warning_behavior,
                "Debug Info Version",
                self._const_int(self.DEBUG_INFO_VERSION),
            ]
        )
        if debuginfo_version not in mflags.operands:
            mflags.add(debuginfo_version)

    def _add_subprogram(self, name, linkagename, line, function, argmap):
        """Emit subprogram metadata"""
        subp = self._di_subprogram(name, linkagename, line, function, argmap)
        self.subprograms.append(subp)
        return subp

    def _add_location(self, line):
        """Emit location metatdaa"""
        loc = self._di_location(line)
        return loc

    @classmethod
    def _const_int(cls, num, bits=32):
        """Util to create constant int in metadata"""
        return ir.IntType(bits)(num)

    @classmethod
    def _const_bool(cls, boolean):
        """Util to create constant boolean in metadata"""
        return ir.IntType(1)(boolean)

    #
    # Helpers to emit the metadata nodes
    #

    def _di_file(self):
        return self.module.add_debug_info(
            "DIFile",
            {
                "directory": os.path.dirname(self.filepath),
                "filename": os.path.basename(self.filepath),
            },
        )

    def _di_compile_unit(self):
        return self.module.add_debug_info(
            "DICompileUnit",
            {
                "language": ir.DIToken("DW_LANG_C_plus_plus"),
                "file": self.difile,
                # Numba has to pretend to be clang to ensure the prologue is skipped
                # correctly in gdb. See:
                # https://sourceware.org/git/?p=binutils-gdb.git;a=blob;f=gdb/amd64-tdep.c;h=e563d369d8cb3eb3c2f732c2fa850ec70ba8d63b;hb=a4b0231e179607e47b1cdf1fe15c5dc25e482fad#l2521
                # Note the "producer_is_llvm" call to specialise the prologue
                # handling, this is defined here:
                # https://sourceware.org/git/?p=binutils-gdb.git;a=blob;f=gdb/producer.c;h=cdfd80d904c09394febd18749bb90359b2d128cc;hb=a4b0231e179607e47b1cdf1fe15c5dc25e482fad#l124
                # and to get a match for this condition the 'producer' must start
                # with "clang ", hence the following...
                "producer": "clang (Numba)",
                "runtimeVersion": 0,
                "isOptimized": config.OPT != 0,
                "emissionKind": ir.DIToken(self.emission_kind),
            },
            is_distinct=True,
        )

    def _di_subroutine_type(self, line, function, argmap):
        # The function call conv needs encoding.
        llfunc = function
        md = []

        for llarg in llfunc.args:
            if not llarg.name.startswith("arg."):
                name = llarg.name.replace(".", "$")  # for gdb to work correctly
                lltype = llarg.type
                size = self.cgctx.get_abi_sizeof(lltype)
                mdtype = self._var_type(lltype, size, datamodel=None)
                md.append(mdtype)

        for name, nbtype in argmap.items():
            name = name.replace(".", "$")  # for gdb to work correctly
            datamodel = self.cgctx.data_model_manager[nbtype]
            lltype = self.cgctx.get_value_type(nbtype)
            size = self.cgctx.get_abi_sizeof(lltype)
            mdtype = self._var_type(lltype, size, datamodel=datamodel)
            md.append(mdtype)

        return self.module.add_debug_info(
            "DISubroutineType",
            {
                "types": self.module.add_metadata(md),
            },
        )

    def _di_subprogram(self, name, linkagename, line, function, argmap):
        return self.module.add_debug_info(
            "DISubprogram",
            {
                "name": name,
                "linkageName": linkagename,
                "scope": self.difile,
                "file": self.difile,
                "line": line,
                "type": self._di_subroutine_type(line, function, argmap),
                "isLocal": False,
                "isDefinition": True,
                "scopeLine": line,
                "isOptimized": config.OPT != 0,
                "unit": self.dicompileunit,
            },
            is_distinct=True,
        )

    def _di_location(self, line):
        return self.module.add_debug_info(
            "DILocation",
            {
                "line": line,
                "column": 1,
                "scope": self.subprograms[-1],
            },
        )


class CUDADIBuilder(DIBuilder):
    def __init__(self, module, filepath, cgctx, directives_only):
        super().__init__(module, filepath, cgctx, directives_only)
        # Cache for local variable metadata type and line deduplication
        self._vartypelinemap = {}

    def _var_type(self, lltype, size, datamodel=None):
        is_bool = False
        is_int_literal = False
        is_grid_group = False
        m = self.module

        if isinstance(lltype, ir.IntType):
            if datamodel is not None:
                name = str(datamodel.fe_type)
                if isinstance(datamodel.fe_type, types.Boolean):
                    is_bool = True
                    if isinstance(datamodel.fe_type, types.BooleanLiteral):
                        name = "bool"
                elif isinstance(datamodel.fe_type, types.Integer):
                    if isinstance(datamodel.fe_type, types.IntegerLiteral):
                        name = f"int{_BYTE_SIZE * size}"
                        is_int_literal = True
                elif isinstance(datamodel.fe_type, GridGroup):
                    is_grid_group = True

        if is_bool or is_int_literal or is_grid_group:
            bitsize = _BYTE_SIZE * size
            # Boolean type workaround until upstream Numba is fixed
            if is_bool:
                ditok = "DW_ATE_boolean"
            elif is_int_literal:
                ditok = "DW_ATE_signed"
            # GridGroup type should use numba.cuda implementation
            elif is_grid_group:
                ditok = "DW_ATE_unsigned"

            return m.add_debug_info(
                "DIBasicType",
                {
                    "name": name,
                    "size": bitsize,
                    "encoding": ir.DIToken(ditok),
                },
            )

        if isinstance(datamodel, UnionModel):
            # UnionModel is handled here to represent polymorphic types
            meta = []
            maxwidth = 0
            for field, model in zip(
                datamodel._fields, datamodel.inner_models()
            ):
                # Ignore the "tag" field, focus on the "payload" field which
                # contains the data types in memory
                if field == "payload":
                    # Store metadata dictionaries to create members later
                    member_metadata_dicts = []

                    for index, mod in enumerate(model.inner_models()):
                        dtype = mod.get_value_type()
                        membersize = self.cgctx.get_abi_sizeof(dtype)
                        basetype = self._var_type(
                            dtype, membersize, datamodel=mod
                        )
                        if isinstance(mod.fe_type, types.Literal):
                            typename = str(mod.fe_type.literal_type)
                        else:
                            typename = str(mod.fe_type)
                        # Use a prefix "_" on type names as field names
                        membername = "_" + typename
                        memberwidth = _BYTE_SIZE * membersize
                        # Build the metadata dictionary
                        metadata_dict = {
                            "tag": ir.DIToken("DW_TAG_member"),
                            "name": membername,
                            "baseType": basetype,
                            # DW_TAG_member size is in bits
                            "size": memberwidth,
                        }
                        if config.CUDA_DEBUG_POLY:
                            # Polymorphic debug info with DW_TAG_variant
                            # extraData depends on llvmlite version
                            if config.CUDA_DEBUG_POLY_USE_TYPED_CONST:
                                metadata_dict["extraData"] = ir.IntType(8)(
                                    index
                                )
                            else:
                                # Use metadata node reference
                                metadata_dict["extraData"] = m.add_metadata(
                                    [ir.IntType(8)(index)]
                                )
                            # Add offset to each variant member
                            # Offset equals the element's own width
                            metadata_dict["offset"] = memberwidth
                        member_metadata_dicts.append(metadata_dict)
                        if memberwidth > maxwidth:
                            maxwidth = memberwidth

                    # Create the member DIDerivedTypes
                    for metadata_dict in member_metadata_dicts:
                        derived_type = m.add_debug_info(
                            "DIDerivedType", metadata_dict
                        )
                        meta.append(derived_type)

            if config.CUDA_DEBUG_POLY:
                # Polymorphic variable debug info generation
                wrapper_struct_size = 2 * maxwidth
                # Generate unique discriminator name based on composite type
                variant_elements_metadata = m.add_metadata(meta)
                discriminator_unique_id = str(id(variant_elements_metadata))
                discriminator_name = f"discriminator-{discriminator_unique_id}"
                discriminator = m.add_debug_info(
                    "DIDerivedType",
                    {
                        "tag": ir.DIToken("DW_TAG_member"),
                        "name": discriminator_name,
                        "baseType": m.add_debug_info(
                            "DIBasicType",
                            {
                                "name": "uint8",
                                "size": _BYTE_SIZE,
                                "encoding": ir.DIToken("DW_ATE_unsigned"),
                            },
                        ),
                        "size": _BYTE_SIZE,
                        "flags": ir.DIToken("DIFlagArtificial"),
                    },
                )
                # Create the final variant_part with actual members
                variant_unique_identifier = discriminator_unique_id
                variant_part_type = m.add_debug_info(
                    "DICompositeType",
                    {
                        "file": self.difile,
                        "tag": ir.DIToken("DW_TAG_variant_part"),
                        "name": "variant_part",
                        "identifier": variant_unique_identifier,
                        "elements": variant_elements_metadata,
                        "size": maxwidth,
                        "discriminator": discriminator,
                    },
                )
                # Create elements metadata for wrapper struct
                elements_metadata = m.add_metadata(
                    [discriminator, variant_part_type]
                )
                unique_identifier = str(id(elements_metadata))
                wrapper_struct = m.add_debug_info(
                    "DICompositeType",
                    {
                        "file": self.difile,
                        "tag": ir.DIToken("DW_TAG_structure_type"),
                        "name": "variant_wrapper_struct",
                        "identifier": unique_identifier,
                        "elements": elements_metadata,
                        "size": wrapper_struct_size,
                    },
                )
                return wrapper_struct
            else:
                fake_union_name = "dbg_poly_union"
                return m.add_debug_info(
                    "DICompositeType",
                    {
                        "file": self.difile,
                        "tag": ir.DIToken("DW_TAG_union_type"),
                        "name": fake_union_name,
                        "identifier": str(lltype),
                        "elements": m.add_metadata(meta),
                        "size": maxwidth,
                    },
                    is_distinct=True,
                )

        # For other cases, use upstream Numba implementation
        return super()._var_type(lltype, size, datamodel=datamodel)

    def _di_subroutine_type(self, line, function, argmap):
        # The function call conv needs encoding.
        llfunc = function
        md = []

        # Create metadata type for return value
        if len(llfunc.args) > 0:
            lltype = llfunc.args[0].type
            size = self.cgctx.get_abi_sizeof(lltype)
            mdtype = self._var_type(lltype, size, datamodel=None)
            md.append(mdtype)

        # Create metadata type for arguments
        for nbtype in argmap.values():
            datamodel = self.cgctx.data_model_manager[nbtype]
            lltype = self.cgctx.get_value_type(nbtype)
            size = self.cgctx.get_abi_sizeof(lltype)
            mdtype = self._var_type(lltype, size, datamodel=datamodel)
            md.append(mdtype)

        return self.module.add_debug_info(
            "DISubroutineType",
            {
                "types": self.module.add_metadata(md),
            },
        )

    def mark_variable(
        self,
        builder,
        allocavalue,
        name,
        lltype,
        size,
        line,
        datamodel=None,
        argidx=None,
    ):
        if name.startswith("$") or "." in name:
            # Do not emit llvm.dbg.declare on user variable alias
            return
        else:
            int_type = (ir.IntType,)
            real_type = ir.FloatType, ir.DoubleType
            if isinstance(lltype, int_type + real_type):
                # For scalar locals we use llvm.dbg.value instead of
                # llvm.dbg.declare, but for scalar *arguments* we still want a
                # stable stack location so they don't get encoded as absolute
                # parameter-space addresses in downstream DWARF.
                if argidx is not None:
                    # NVVM has been observed to crash on some boolean-parameter
                    # debug.declare patterns - use dbg.value for these instead.
                    if datamodel is not None and isinstance(
                        datamodel.fe_type, types.Boolean
                    ):
                        return
                    return super().mark_variable(
                        builder,
                        allocavalue,
                        name,
                        lltype,
                        size,
                        line,
                        datamodel,
                        argidx,
                    )
                return
            else:
                return super().mark_variable(
                    builder,
                    allocavalue,
                    name,
                    lltype,
                    size,
                    line,
                    datamodel,
                    argidx,
                )

    def update_variable(
        self,
        builder,
        value,
        name,
        lltype,
        size,
        line,
        datamodel=None,
        argidx=None,
    ):
        m = self.module
        fnty = ir.FunctionType(ir.VoidType(), [ir.MetaDataType()] * 3)
        decl = cgutils.get_or_insert_function(m, fnty, "llvm.dbg.value")

        mdtype = self._var_type(lltype, size, datamodel)
        index = name.find(".")
        if index >= 0:
            name = name[:index]
        # Merge DILocalVariable nodes with same name and type but different
        # lines. Use the cached [(name, type) -> line] info to deduplicate
        # metadata. Use the lltype as part of key.
        key = (name, lltype)
        if key in self._vartypelinemap:
            line = self._vartypelinemap[key]
        else:
            self._vartypelinemap[key] = line
        arg_index = 0 if argidx is None else argidx
        mdlocalvar = m.add_debug_info(
            "DILocalVariable",
            {
                "name": name,
                "arg": arg_index,
                "scope": self.subprograms[-1],
                "file": self.difile,
                "line": line,
                "type": mdtype,
            },
        )
        mdexpr = m.add_debug_info("DIExpression", {})

        return builder.call(decl, [value, mdlocalvar, mdexpr])



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause

from collections import defaultdict

from numba.cuda import config


class Rewrite:
    """Defines the abstract base class for Numba rewrites."""

    def __init__(self, state=None):
        """Constructor for the Rewrite class."""
        pass

    def match(self, func_ir, block, typemap, calltypes) -> bool:
        """Overload this method to check an IR block for matching terms in the
        rewrite.
        """
        return False

    def apply(self):
        """Overload this method to return a rewritten IR basic block when a
        match has been found.
        """
        raise NotImplementedError("Abstract Rewrite.apply() called!")


class RewriteRegistry:
    """Defines a registry for Numba rewrites."""

    _kinds = frozenset(["before-inference", "after-inference"])

    def __init__(self):
        """Constructor for the rewrite registry.  Initializes the rewrites
        member to an empty list.
        """
        self.rewrites = defaultdict(list)

    def register(self, kind):
        """
        Decorator adding a subclass of Rewrite to the registry for
        the given *kind*.
        """
        if kind not in self._kinds:
            raise KeyError("invalid kind %r" % (kind,))

        def do_register(rewrite_cls):
            if not issubclass(rewrite_cls, Rewrite):
                raise TypeError(
                    "{0} is not a subclass of Rewrite".format(rewrite_cls)
                )
            self.rewrites[kind].append(rewrite_cls)
            return rewrite_cls

        return do_register

    def apply(self, kind, state):
        """Given a pipeline and a dictionary of basic blocks, exhaustively
        attempt to apply all registered rewrites to all basic blocks.
        """
        assert kind in self._kinds
        blocks = state.func_ir.blocks
        old_blocks = blocks.copy()
        for rewrite_cls in self.rewrites[kind]:
            # Exhaustively apply a rewrite until it stops matching.
            rewrite = rewrite_cls(state)
            work_list = list(blocks.items())
            while work_list:
                key, block = work_list.pop()
                matches = rewrite.match(
                    state.func_ir, block, state.typemap, state.calltypes
                )
                if matches:
                    if config.DEBUG or config.DUMP_IR:
                        print("_" * 70)
                        print("REWRITING (%s):" % rewrite_cls.__name__)
                        block.dump()
                        print("_" * 60)
                    new_block = rewrite.apply()
                    blocks[key] = new_block
                    work_list.append((key, new_block))
                    if config.DEBUG or config.DUMP_IR:
                        new_block.dump()
                        print("_" * 70)
        # If any blocks were changed, perform a sanity check.
        for key, block in blocks.items():
            if block != old_blocks[key]:
                block.verify()

        # Some passes, e.g. _inline_const_arraycall are known to occasionally
        # do invalid things WRT ir.Del, others, e.g. RewriteArrayExprs do valid
        # things with ir.Del, but the placement is not optimal. The lines below
        # fix-up the IR so that ref counts are valid and optimally placed,
        # see #4093 for context. This has to be run here opposed to in
        # apply() as the CFG needs computing so full IR is needed.
        from numba.cuda.core import postproc

        post_proc = postproc.PostProcessor(state.func_ir)
        post_proc.run()


rewrite_registry = RewriteRegistry()
register_rewrite = rewrite_registry.register



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause

"""
Function descriptors.
"""

from collections import defaultdict
import importlib

from numba.cuda import types, cgutils
from numba.cuda import itanium_mangler
from numba.cuda.utils import _dynamic_modname, _dynamic_module


def default_mangler(name, argtypes, *, abi_tags=(), uid=None):
    return itanium_mangler.mangle(name, argtypes, abi_tags=abi_tags, uid=uid)


def qualifying_prefix(modname, qualname):
    """
    Returns a new string that is used for the first half of the mangled name.
    """
    # XXX choose a different convention for object mode
    return "{}.{}".format(modname, qualname) if modname else qualname


class FunctionDescriptor:
    """
    Base class for function descriptors: an object used to carry
    useful metadata about a natively callable function.

    Note that while `FunctionIdentity` denotes a Python function
    which is being concretely compiled by Numba, `FunctionDescriptor`
    may be more "abstract".
    """

    __slots__ = (
        "native",
        "modname",
        "qualname",
        "doc",
        "typemap",
        "calltypes",
        "args",
        "kws",
        "restype",
        "argtypes",
        "mangled_name",
        "unique_name",
        "env_name",
        "global_dict",
        "inline",
        "noalias",
        "abi_tags",
        "uid",
        "call_conv",
        "abi_info",
    )

    def __init__(
        self,
        native,
        modname,
        qualname,
        unique_name,
        doc,
        typemap,
        restype,
        calltypes,
        args,
        kws,
        mangler=None,
        argtypes=None,
        inline=False,
        noalias=False,
        env_name=None,
        global_dict=None,
        abi_tags=(),
        uid=None,
        call_conv=None,
        abi_info=None,
    ):
        self.native = native
        self.modname = modname
        self.global_dict = global_dict
        self.qualname = qualname
        self.unique_name = unique_name
        self.doc = doc
        # XXX typemap and calltypes should be on the compile result,
        # not the FunctionDescriptor
        self.typemap = typemap
        self.calltypes = calltypes
        self.args = args
        self.kws = kws
        self.restype = restype
        # Argument types
        if argtypes is not None:
            assert isinstance(argtypes, tuple), argtypes
            self.argtypes = argtypes
        else:
            # Get argument types from the type inference result
            # (note the "arg.FOO" convention as used in typeinfer
            self.argtypes = tuple(self.typemap["arg." + a] for a in args)
        mangler = default_mangler if mangler is None else mangler
        # The mangled name *must* be unique, else the wrong function can
        # be chosen at link time.
        qualprefix = qualifying_prefix(self.modname, self.qualname)
        self.uid = uid

        if abi_info is not None and "abi_name" in abi_info:
            self.mangled_name = abi_info["abi_name"]
        else:
            self.mangled_name = mangler(
                qualprefix,
                self.argtypes,
                abi_tags=abi_tags,
                uid=uid,
            )

        if env_name is None:
            env_name = mangler(
                ".NumbaEnv.{}".format(qualprefix),
                self.argtypes,
                abi_tags=abi_tags,
                uid=uid,
            )
        self.env_name = env_name
        self.inline = inline
        self.noalias = noalias
        self.abi_tags = abi_tags
        self.call_conv = call_conv
        self.abi_info = abi_info

    def lookup_globals(self):
        """
        Return the global dictionary of the function.
        It may not match the Module's globals if the function is created
        dynamically (i.e. exec)
        """
        return self.global_dict or self.lookup_module().__dict__

    def lookup_module(self):
        """
        Return the module in which this function is supposed to exist.
        This may be a dummy module if the function was dynamically
        generated or the module can't be found.
        """
        if self.modname == _dynamic_modname:
            return _dynamic_module
        else:
            try:
                # ensure module exist
                return importlib.import_module(self.modname)
            except ImportError:
                return _dynamic_module

    def lookup_function(self):
        """
        Return the original function object described by this object.
        """
        return getattr(self.lookup_module(), self.qualname)

    @property
    def llvm_func_name(self):
        """
        The LLVM-registered name for the raw function.
        """
        return self.mangled_name

    @property
    def llvm_cpython_wrapper_name(self):
        """
        The LLVM-registered name for a CPython-compatible wrapper of the
        raw function (i.e. a PyCFunctionWithKeywords).
        """
        return itanium_mangler.prepend_namespace(
            self.mangled_name, ns="cpython"
        )

    @property
    def llvm_cfunc_wrapper_name(self):
        """
        The LLVM-registered name for a C-compatible wrapper of the
        raw function.
        """
        return "cfunc." + self.mangled_name

    def __repr__(self):
        return "<function descriptor %r>" % (self.unique_name)

    @classmethod
    def _get_function_info(cls, func_ir):
        """
        Returns
        -------
        qualname, unique_name, modname, doc, args, kws, globals

        ``unique_name`` must be a unique name.
        """
        func = func_ir.func_id.func
        qualname = func_ir.func_id.func_qualname
        # XXX to func_id
        modname = func.__module__
        doc = func.__doc__ or ""
        args = tuple(func_ir.arg_names)
        kws = ()  # TODO
        global_dict = None

        if modname is None:
            # Dynamically generated function.
            modname = _dynamic_modname
            # Retain a reference to the dictionary of the function.
            # This disables caching, serialization and pickling.
            global_dict = func_ir.func_id.func.__globals__

        unique_name = func_ir.func_id.unique_name

        return qualname, unique_name, modname, doc, args, kws, global_dict

    @classmethod
    def _from_python_function(
        cls,
        func_ir,
        typemap,
        restype,
        calltypes,
        native,
        mangler=None,
        inline=False,
        noalias=False,
        abi_tags=(),
        call_conv=None,
        abi_info=None,
    ):
        (
            qualname,
            unique_name,
            modname,
            doc,
            args,
            kws,
            global_dict,
        ) = cls._get_function_info(func_ir)

        self = cls(
            native,
            modname,
            qualname,
            unique_name,
            doc,
            typemap,
            restype,
            calltypes,
            args,
            kws,
            mangler=mangler,
            inline=inline,
            noalias=noalias,
            global_dict=global_dict,
            abi_tags=abi_tags,
            uid=func_ir.func_id.unique_id,
            call_conv=call_conv,
            abi_info=abi_info,
        )
        return self

    def declare_function(self, module):
        fnty = self.call_conv.get_function_type(self.restype, self.argtypes)
        fn = cgutils.get_or_insert_function(module, fnty, self.mangled_name)
        self.call_conv.decorate_function(
            fn, self.args, self.argtypes, noalias=self.noalias
        )
        if self.inline:
            fn.attributes.add("alwaysinline")
            # alwaysinline overrides optnone
            fn.attributes.discard("noinline")
            fn.attributes.discard("optnone")
        return fn


class PythonFunctionDescriptor(FunctionDescriptor):
    """
    A FunctionDescriptor subclass for Numba-compiled functions.
    """

    __slots__ = ()

    @classmethod
    def from_specialized_function(
        cls,
        func_ir,
        typemap,
        restype,
        calltypes,
        mangler,
        inline,
        noalias,
        abi_tags,
        call_conv,
        abi_info,
    ):
        """
        Build a FunctionDescriptor for a given specialization of a Python
        function (in nopython mode).
        """
        return cls._from_python_function(
            func_ir,
            typemap,
            restype,
            calltypes,
            native=True,
            mangler=mangler,
            inline=inline,
            noalias=noalias,
            abi_tags=abi_tags,
            call_conv=call_conv,
            abi_info=abi_info,
        )

    @classmethod
    def from_object_mode_function(cls, func_ir):
        """
        Build a FunctionDescriptor for an object mode variant of a Python
        function.
        """
        typemap = defaultdict(lambda: types.pyobject)
        calltypes = typemap.copy()
        restype = types.pyobject
        return cls._from_python_function(
            func_ir, typemap, restype, calltypes, native=False
        )


class ExternalFunctionDescriptor(FunctionDescriptor):
    """
    A FunctionDescriptor subclass for opaque external functions
    (e.g. raw C functions).
    """

    __slots__ = ()

    def __init__(self, name, restype, argtypes, call_conv):
        args = ["arg%d" % i for i in range(len(argtypes))]

        def mangler(a, x, abi_tags, uid=None):
            return a

        super().__init__(
            native=True,
            modname=None,
            qualname=name,
            unique_name=name,
            doc="",
            typemap=None,
            restype=restype,
            calltypes=None,
            args=args,
            kws=None,
            mangler=mangler,
            argtypes=argtypes,
            call_conv=call_conv,
        )



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause

"""
Implementation of some CFFI functions
"""

from numba.cuda.core.imputils import Registry
from numba.cuda import types

registry = Registry("cffiimpl")


@registry.lower("ffi.from_buffer", types.Buffer)
def from_buffer(context, builder, sig, args):
    assert len(sig.args) == 1
    assert len(args) == 1
    [fromty] = sig.args
    [val] = args
    # Type inference should have prevented passing a buffer from an
    # array to a pointer of the wrong type
    assert fromty.dtype == sig.return_type.dtype
    ary = context.make_array(fromty)(context, builder, val)
    return ary.data



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause
from numba.cuda._internal.cuda_fp8 import (
    typing_registry,
    target_registry,
)


__all__ = [
    "typing_registry",
    "target_registry",
]



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause

"""
Implement code coverage support.

Currently contains logic to extend ``coverage`` with lines covered by the
compiler.
"""

from typing import Optional, Sequence, Callable
from abc import ABC, abstractmethod

from numba.cuda.core import ir
from numba.cuda import config

_the_registry: Callable[[], Optional["NotifyLocBase"]] = []


def get_registered_loc_notify() -> Sequence["NotifyLocBase"]:
    """
    Returns a list of the registered NotifyLocBase instances.
    """
    if hasattr(config, "JIT_COVERAGE") and not config.JIT_COVERAGE:
        # Coverage disabled.
        return []
    return list(
        filter(
            lambda x: x is not None, (factory() for factory in _the_registry)
        )
    )


class NotifyLocBase(ABC):
    """Interface for notifying visiting of a ``numba.cuda.core.ir.Loc``."""

    @abstractmethod
    def notify(self, loc: ir.Loc) -> None:
        pass

    @abstractmethod
    def close(self) -> None:
        pass



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause

"""
The Device Array API is not implemented in the simulator. This module provides
stubs to allow tests to import correctly.
"""

from contextlib import contextmanager
from numba.cuda.np.numpy_support import numpy_version
from numba.cuda.np import numpy_support
from numba.cuda import types

import numpy as np


DeviceRecord = None
from_record_like = None


errmsg_contiguous_buffer = (
    "Array contains non-contiguous buffer and cannot "
    "be transferred as a single memory region. Please "
    "ensure contiguous buffer with numpy "
    ".ascontiguousarray()"
)


class FakeShape(tuple):
    """
    The FakeShape class is used to provide a shape which does not allow negative
    indexing, similar to the shape in CUDA Python. (Numpy shape arrays allow
    negative indexing)
    """

    def __getitem__(self, k):
        if isinstance(k, int) and k < 0:
            raise IndexError("tuple index out of range")
        return super().__getitem__(k)


class FakeWithinKernelCUDAArray:
    """
    Created to emulate the behavior of arrays within kernels, where either
    array.item or array['item'] is valid (that is, give all structured
    arrays `numpy.recarray`-like semantics). This behaviour does not follow
    the semantics of Python and NumPy with non-jitted code, and will be
    deprecated and removed.
    """

    def __init__(self, item):
        assert isinstance(item, FakeCUDAArray)
        self.__dict__["_item"] = item

    def __wrap_if_fake(self, item):
        if isinstance(item, FakeCUDAArray):
            return FakeWithinKernelCUDAArray(item)
        else:
            return item

    def __getattr__(self, attrname):
        if (value := getattr(self._item, attrname, None)) is not None:
            return self.__wrap_if_fake(value)
        else:
            return self.__wrap_if_fake(self._item.__getitem__(attrname))

    def __setattr__(self, nm, val):
        self._item.__setitem__(nm, val)

    def __getitem__(self, idx):
        return self.__wrap_if_fake(self._item.__getitem__(idx))

    def __setitem__(self, idx, val):
        self._item.__setitem__(idx, val)

    def __len__(self):
        return len(self._item)

    def __array_ufunc__(self, ufunc, method, *args, **kwargs):
        # ufuncs can only be called directly on instances of numpy.ndarray (not
        # things that implement its interfaces, like the FakeCUDAArray or
        # FakeWithinKernelCUDAArray). For other objects, __array_ufunc__ is
        # called when they are arguments to ufuncs, to provide an opportunity
        # to somehow implement the ufunc. Since the FakeWithinKernelCUDAArray
        # is just a thin wrapper over an ndarray, we can implement all ufuncs
        # by passing the underlying ndarrays to a call to the intended ufunc.
        call = getattr(ufunc, method)

        def convert_fakes(obj):
            if isinstance(obj, FakeWithinKernelCUDAArray):
                obj = obj._item._ary

            return obj

        out = kwargs.get("out")
        if out:
            kwargs["out"] = tuple(convert_fakes(o) for o in out)
        args = tuple(convert_fakes(a) for a in args)
        return call(*args, **kwargs)


class FakeCUDAArray:
    """
    Implements the interface of a DeviceArray/DeviceRecord, but mostly just
    wraps a NumPy array.
    """

    __cuda_ndarray__ = True  # There must be gpu_data attribute

    def __init__(self, ary, stream=0):
        self._ary = ary
        self.stream = stream

    @property
    def _numba_type_(self):
        """
        Magic attribute expected by Numba to get the numba type that
        represents this object.
        """
        broadcast = 0 in self.strides
        if self.is_c_contiguous() and not broadcast:
            layout = "C"
        elif self.is_f_contiguous() and not broadcast:
            layout = "F"
        else:
            layout = "A"

        dtype = numpy_support.from_dtype(self._ary.dtype)
        return types.Array(dtype, self._ary.ndim, layout)

    @property
    def alloc_size(self):
        return self._ary.nbytes

    @property
    def nbytes(self):
        # return nbytes -- FakeCUDAArray is a wrapper around NumPy
        return self._ary.nbytes

    def __getattr__(self, attrname):
        try:
            attr = getattr(self._ary, attrname)
            return attr
        except AttributeError as e:
            msg = "Wrapped array has no attribute '%s'" % attrname
            raise AttributeError(msg) from e

    def bind(self, stream=0):
        return FakeCUDAArray(self._ary, stream)

    @property
    def T(self):
        return self.transpose()

    def transpose(self, axes=None):
        return FakeCUDAArray(np.transpose(self._ary, axes=axes))

    def __getitem__(self, idx):
        ret = self._ary.__getitem__(idx)
        if type(ret) not in [np.ndarray, np.void]:
            return ret
        else:
            return FakeCUDAArray(ret, stream=self.stream)

    def __setitem__(self, idx, val):
        return self._ary.__setitem__(idx, val)

    def copy_to_host(self, ary=None, stream=0):
        if ary is None:
            ary = np.empty_like(self._ary)
        else:
            check_array_compatibility(self, ary)
        np.copyto(ary, self._ary)
        return ary

    def copy_to_device(self, ary, stream=0):
        """
        Copy from the provided array into this array.

        This may be less forgiving than the CUDA Python implementation, which
        will copy data up to the length of the smallest of the two arrays,
        whereas this expects the size of the arrays to be equal.
        """
        sentry_contiguous(self)
        self_core, ary_core = array_core(self), array_core(ary)
        if isinstance(ary, FakeCUDAArray):
            sentry_contiguous(ary)
            check_array_compatibility(self_core, ary_core)
        else:
            ary_core = np.array(
                ary_core,
                order="C" if self_core.flags["C_CONTIGUOUS"] else "F",
                subok=True,
                copy=False if numpy_version < (2, 0) else None,
            )
            check_array_compatibility(self_core, ary_core)
        np.copyto(self_core._ary, ary_core)

    @property
    def shape(self):
        return FakeShape(self._ary.shape)

    def ravel(self, *args, **kwargs):
        return FakeCUDAArray(self._ary.ravel(*args, **kwargs))

    def reshape(self, *args, **kwargs):
        return FakeCUDAArray(self._ary.reshape(*args, **kwargs))

    def view(self, *args, **kwargs):
        return FakeCUDAArray(self._ary.view(*args, **kwargs))

    def is_c_contiguous(self):
        return self._ary.flags.c_contiguous

    def is_f_contiguous(self):
        return self._ary.flags.f_contiguous

    def __str__(self):
        return str(self._ary)

    def __repr__(self):
        return repr(self._ary)

    def __len__(self):
        return len(self._ary)

    # TODO: Add inplace, bitwise, unary magic methods
    #  (or maybe inherit this class from numpy)?
    def __eq__(self, other):
        return FakeCUDAArray(self._ary == other)

    def __ne__(self, other):
        return FakeCUDAArray(self._ary != other)

    def __lt__(self, other):
        return FakeCUDAArray(self._ary < other)

    def __le__(self, other):
        return FakeCUDAArray(self._ary <= other)

    def __gt__(self, other):
        return FakeCUDAArray(self._ary > other)

    def __ge__(self, other):
        return FakeCUDAArray(self._ary >= other)

    def __add__(self, other):
        return FakeCUDAArray(self._ary + other)

    def __sub__(self, other):
        return FakeCUDAArray(self._ary - other)

    def __mul__(self, other):
        return FakeCUDAArray(self._ary * other)

    def __floordiv__(self, other):
        return FakeCUDAArray(self._ary // other)

    def __truediv__(self, other):
        return FakeCUDAArray(self._ary / other)

    def __mod__(self, other):
        return FakeCUDAArray(self._ary % other)

    def __pow__(self, other):
        return FakeCUDAArray(self._ary**other)

    def split(self, section, stream=0):
        return [
            FakeCUDAArray(a)
            for a in np.split(self._ary, range(section, len(self), section))
        ]


def array_core(ary):
    """
    Extract the repeated core of a broadcast array.

    Broadcast arrays are by definition non-contiguous due to repeated
    dimensions, i.e., dimensions with stride 0. In order to ascertain memory
    contiguity and copy the underlying data from such arrays, we must create
    a view without the repeated dimensions.

    """
    if not ary.strides or not ary.size:
        return ary
    core_index = []
    for stride in ary.strides:
        core_index.append(0 if stride == 0 else slice(None))
    return ary[tuple(core_index)]


def is_contiguous(ary):
    """
    Returns True iff `ary` is C-style contiguous while ignoring
    broadcasted and 1-sized dimensions.
    As opposed to array_core(), it does not call require_context(),
    which can be quite expensive.
    """
    size = ary.dtype.itemsize
    for shape, stride in zip(reversed(ary.shape), reversed(ary.strides)):
        if shape > 1 and stride != 0:
            if size != stride:
                return False
            size *= shape
    return True


def sentry_contiguous(ary):
    core = array_core(ary)
    if not core.flags["C_CONTIGUOUS"] and not core.flags["F_CONTIGUOUS"]:
        raise ValueError(errmsg_contiguous_buffer)


def check_array_compatibility(ary1, ary2):
    ary1sq, ary2sq = ary1.squeeze(), ary2.squeeze()
    if ary1.dtype != ary2.dtype:
        raise TypeError(
            "incompatible dtype: %s vs. %s" % (ary1.dtype, ary2.dtype)
        )
    if ary1sq.shape != ary2sq.shape:
        raise ValueError(
            "incompatible shape: %s vs. %s" % (ary1.shape, ary2.shape)
        )
    if ary1sq.strides != ary2sq.strides:
        raise ValueError(
            "incompatible strides: %s vs. %s" % (ary1.strides, ary2.strides)
        )


def to_device(ary, stream=0, copy=True, to=None):
    ary = np.array(
        ary, copy=False if numpy_version < (2, 0) else None, subok=True
    )
    sentry_contiguous(ary)
    if to is None:
        buffer_dtype = np.int64 if ary.dtype.char in "Mm" else ary.dtype
        return FakeCUDAArray(
            np.ndarray(
                buffer=np.copy(array_core(ary)).view(buffer_dtype),
                dtype=ary.dtype,
                shape=ary.shape,
                strides=ary.strides,
            ).view(type=type(ary)),
        )
    else:
        to.copy_to_device(ary, stream=stream)


@contextmanager
def pinned(arg):
    yield


def mapped_array(*args, **kwargs):
    for unused_arg in ("portable", "wc"):
        if unused_arg in kwargs:
            kwargs.pop(unused_arg)
    return device_array(*args, **kwargs)


def pinned_array(shape, dtype=np.float64, strides=None, order="C"):
    return np.ndarray(shape=shape, strides=strides, dtype=dtype, order=order)


def managed_array(shape, dtype=np.float64, strides=None, order="C"):
    return np.ndarray(shape=shape, strides=strides, dtype=dtype, order=order)


def device_array(*args, **kwargs):
    stream = kwargs.pop("stream") if "stream" in kwargs else 0
    return FakeCUDAArray(np.ndarray(*args, **kwargs), stream=stream)


def _contiguous_strides_like_array(ary):
    """
    Given an array, compute strides for a new contiguous array of the same
    shape.
    """
    # Don't recompute strides if the default strides will be sufficient to
    # create a contiguous array.
    if ary.flags["C_CONTIGUOUS"] or ary.flags["F_CONTIGUOUS"] or ary.ndim <= 1:
        return None

    # Otherwise, we need to compute new strides using an algorithm adapted from
    # NumPy v1.17.4's PyArray_NewLikeArrayWithShape in
    # core/src/multiarray/ctors.c. We permute the strides in ascending order
    # then compute the stride for the dimensions with the same permutation.

    # Stride permutation. E.g. a stride array (4, -2, 12) becomes
    # [(1, -2), (0, 4), (2, 12)]
    strideperm = [x for x in enumerate(ary.strides)]
    strideperm.sort(key=lambda x: x[1])

    # Compute new strides using permutation
    strides = [0] * len(ary.strides)
    stride = ary.dtype.itemsize
    for i_perm, _ in strideperm:
        strides[i_perm] = stride
        stride *= ary.shape[i_perm]
    return tuple(strides)


def _order_like_array(ary):
    if ary.flags["F_CONTIGUOUS"] and not ary.flags["C_CONTIGUOUS"]:
        return "F"
    else:
        return "C"


def device_array_like(ary, stream=0):
    strides = _contiguous_strides_like_array(ary)
    order = _order_like_array(ary)
    return device_array(
        shape=ary.shape, dtype=ary.dtype, strides=strides, order=order
    )


def pinned_array_like(ary):
    strides = _contiguous_strides_like_array(ary)
    order = _order_like_array(ary)
    return pinned_array(
        shape=ary.shape, dtype=ary.dtype, strides=strides, order=order
    )


def auto_device(ary, stream=0, copy=True):
    if isinstance(ary, FakeCUDAArray):
        return ary, False

    if not isinstance(ary, np.void):
        ary = np.array(
            ary, copy=False if numpy_version < (2, 0) else None, subok=True
        )
    return to_device(ary, stream, copy), True


def is_cuda_ndarray(obj):
    "Check if an object is a CUDA ndarray"
    return getattr(obj, "__cuda_ndarray__", False)


def verify_cuda_ndarray_interface(obj):
    "Verify the CUDA ndarray interface for an obj"
    require_cuda_ndarray(obj)

    def requires_attr(attr, typ):
        if not hasattr(obj, attr):
            raise AttributeError(attr)
        if not isinstance(getattr(obj, attr), typ):
            raise AttributeError("%s must be of type %s" % (attr, typ))

    requires_attr("shape", tuple)
    requires_attr("strides", tuple)
    requires_attr("dtype", np.dtype)
    requires_attr("size", int)


def require_cuda_ndarray(obj):
    "Raises ValueError is is_cuda_ndarray(obj) evaluates False"
    if not is_cuda_ndarray(obj):
        raise ValueError("require an cuda ndarray object")



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause

"""
Defines CUDA Options for use in the CUDA target
"""

from abc import ABCMeta, abstractmethod


class AbstractOptionValue(metaclass=ABCMeta):
    """Abstract base class for custom option values."""

    @abstractmethod
    def encode(self) -> str:
        """Returns an encoding of the values"""
        ...

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.encode()})"


class FastMathOptions(AbstractOptionValue):
    """
    Options for controlling fast math optimization.
    """

    def __init__(self, value):
        # https://releases.llvm.org/7.0.0/docs/LangRef.html#fast-math-flags
        valid_flags = {
            "fast",
            "nnan",
            "ninf",
            "nsz",
            "arcp",
            "contract",
            "afn",
            "reassoc",
        }

        if isinstance(value, FastMathOptions):
            self.flags = value.flags.copy()
        elif value is True:
            self.flags = {"fast"}
        elif value is False:
            self.flags = set()
        elif isinstance(value, set):
            invalid = value - valid_flags
            if invalid:
                raise ValueError("Unrecognized fastmath flags: %s" % invalid)
            self.flags = value
        elif isinstance(value, dict):
            invalid = set(value.keys()) - valid_flags
            if invalid:
                raise ValueError("Unrecognized fastmath flags: %s" % invalid)
            self.flags = {v for v, enable in value.items() if enable}
        else:
            msg = "Expected fastmath option(s) to be either a bool, dict or set"
            raise ValueError(msg)

    def __bool__(self):
        return bool(self.flags)

    __nonzero__ = __bool__

    def encode(self) -> str:
        return str(self.flags)

    def __eq__(self, other):
        if type(other) is type(self):
            return self.flags == other.flags
        return NotImplemented


class ParallelOptions(AbstractOptionValue):
    """
    Options for controlling auto parallelization.
    """

    __slots__ = (
        "enabled",
        "comprehension",
        "reduction",
        "inplace_binop",
        "setitem",
        "numpy",
        "stencil",
        "fusion",
        "prange",
    )

    def __init__(self, value):
        if isinstance(value, bool):
            self.enabled = value
            self.comprehension = value
            self.reduction = value
            self.inplace_binop = value
            self.setitem = value
            self.numpy = value
            self.stencil = value
            self.fusion = value
            self.prange = value
        elif isinstance(value, dict):
            self.enabled = True
            self.comprehension = value.pop("comprehension", True)
            self.reduction = value.pop("reduction", True)
            self.inplace_binop = value.pop("inplace_binop", True)
            self.setitem = value.pop("setitem", True)
            self.numpy = value.pop("numpy", True)
            self.stencil = value.pop("stencil", True)
            self.fusion = value.pop("fusion", True)
            self.prange = value.pop("prange", True)
            if value:
                msg = "Unrecognized parallel options: %s" % value.keys()
                raise NameError(msg)
        elif isinstance(value, ParallelOptions):
            self.enabled = value.enabled
            self.comprehension = value.comprehension
            self.reduction = value.reduction
            self.inplace_binop = value.inplace_binop
            self.setitem = value.setitem
            self.numpy = value.numpy
            self.stencil = value.stencil
            self.fusion = value.fusion
            self.prange = value.prange
        else:
            msg = "Expect parallel option to be either a bool or a dict"
            raise ValueError(msg)

    def _get_values(self):
        """Get values as dictionary."""
        return {k: getattr(self, k) for k in self.__slots__}

    def __eq__(self, other):
        if type(other) is type(self):
            return self._get_values() == other._get_values()
        return NotImplemented

    def encode(self) -> str:
        return ", ".join(f"{k}={v}" for k, v in self._get_values().items())


class InlineOptions(AbstractOptionValue):
    """
    Options for controlling inlining
    """

    def __init__(self, value):
        ok = False
        if isinstance(value, str):
            if value in ("always", "never"):
                ok = True
        else:
            ok = callable(value)

        if ok:
            self._inline = value
        else:
            msg = (
                "kwarg 'inline' must be one of the strings 'always' or "
                "'never', or it can be a callable that returns True/False. "
                "Found value %s" % value
            )
            raise ValueError(msg)

    @property
    def is_never_inline(self):
        """
        True if never inline
        """
        return self._inline == "never"

    @property
    def is_always_inline(self):
        """
        True if always inline
        """
        return self._inline == "always"

    @property
    def has_cost_model(self):
        """
        True if a cost model is provided
        """
        return not (self.is_always_inline or self.is_never_inline)

    @property
    def value(self):
        """
        The raw value
        """
        return self._inline

    def __eq__(self, other):
        if type(other) is type(self):
            return self.value == other.value
        return NotImplemented

    def encode(self) -> str:
        return repr(self._inline)


class TargetOptions:
    """Target options maps user options from decorators to the
    ``numba.cuda.core.compiler.Flags`` used by lowering and target context.
    """

    class Mapping:
        def __init__(self, flag_name, apply=lambda x: x):
            self.flag_name = flag_name
            self.apply = apply

    def finalize(self, flags, options):
        """Subclasses can override this method to make target specific
        customizations of default flags.

        Parameters
        ----------
        flags : Flags
        options : dict
        """
        pass

    @classmethod
    def parse_as_flags(cls, flags, options):
        """Parse target options defined in ``options`` and set ``flags``
        accordingly.

        Parameters
        ----------
        flags : Flags
        options : dict
        """
        opt = cls()
        opt._apply(flags, options)
        opt.finalize(flags, options)
        return flags

    def _apply(self, flags, options):
        # Find all Mapping instances in the class
        mappings = {}
        cls = type(self)
        for k in dir(cls):
            v = getattr(cls, k)
            if isinstance(v, cls.Mapping):
                mappings[k] = v

        used = set()
        for k, mapping in mappings.items():
            if k in options:
                v = mapping.apply(options[k])
                setattr(flags, mapping.flag_name, v)
                used.add(k)

        unused = set(options) - used
        if unused:
            # Unread options?
            m = (
                f"Unrecognized options: {unused}. "
                f"Known options are {mappings.keys()}"
            )
            raise KeyError(m)



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause

"""
Helper classes / mixins for defining types.
"""

from .abstract import ArrayCompatible, Dummy, IterableType, IteratorType
from numba.cuda.core.errors import NumbaTypeError, NumbaValueError


class Opaque(Dummy):
    """
    A type that is a opaque pointer.
    """


class SimpleIterableType(IterableType):
    def __init__(self, name, iterator_type):
        self._iterator_type = iterator_type
        super().__init__(name)

    @property
    def iterator_type(self):
        return self._iterator_type


class SimpleIteratorType(IteratorType):
    def __init__(self, name, yield_type):
        self._yield_type = yield_type
        super().__init__(name)

    @property
    def yield_type(self):
        return self._yield_type


class Buffer(IterableType, ArrayCompatible):
    """
    Type class for objects providing the buffer protocol.
    Derived classes exist for more specific cases.
    """

    mutable = True
    slice_is_copy = False
    aligned = True

    # CS and FS are not reserved for inner contig but strided
    LAYOUTS = frozenset(["C", "F", "CS", "FS", "A"])

    def __init__(self, dtype, ndim, layout, readonly=False, name=None):
        from .misc import unliteral

        if isinstance(dtype, Buffer):
            msg = (
                "The dtype of a Buffer type cannot itself be a Buffer type, "
                "this is unsupported behaviour."
                "\nThe dtype requested for the unsupported Buffer was: {}."
            )
            raise NumbaTypeError(msg.format(dtype))
        if layout not in self.LAYOUTS:
            raise NumbaValueError("Invalid layout '%s'" % layout)
        self.dtype = unliteral(dtype)
        self.ndim = ndim
        self.layout = layout
        if readonly:
            self.mutable = False
        if name is None:
            type_name = self.__class__.__name__.lower()
            if readonly:
                type_name = "readonly %s" % type_name
            name = "%s(%s, %sd, %s)" % (type_name, dtype, ndim, layout)
        super().__init__(name)

    @property
    def iterator_type(self):
        from .iterators import ArrayIterator

        return ArrayIterator(self)

    @property
    def as_array(self):
        return self

    def copy(self, dtype=None, ndim=None, layout=None):
        if dtype is None:
            dtype = self.dtype
        if ndim is None:
            ndim = self.ndim
        if layout is None:
            layout = self.layout
        return self.__class__(
            dtype=dtype, ndim=ndim, layout=layout, readonly=not self.mutable
        )

    @property
    def key(self):
        return self.dtype, self.ndim, self.layout, self.mutable

    @property
    def is_c_contig(self):
        return self.layout == "C" or (self.ndim <= 1 and self.layout in "CF")

    @property
    def is_f_contig(self):
        return self.layout == "F" or (self.ndim <= 1 and self.layout in "CF")

    @property
    def is_contig(self):
        return self.layout in "CF"



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause

"""
Itanium CXX ABI Mangler

Reference: https://itanium-cxx-abi.github.io/cxx-abi/abi.html

The basics of the mangling scheme.

We are hijacking the CXX mangling scheme for our use.  We map Python modules
into CXX namespace.  A `module1.submodule2.foo` is mapped to
`module1::submodule2::foo`.   For parameterized numba types, we treat them as
templated types; for example, `array(int64, 1d, C)` becomes an
`array<int64, 1, C>`.

All mangled names are prefixed with "_Z".  It is followed by the name of the
entity.  A name contains one or more identifiers.  Each identifier is encoded
as "<num of char><name>".   If the name is namespaced and, therefore,
has multiple identifiers, the entire name is encoded as "N<name>E".

For functions, arguments types follow.  There are condensed encodings for basic
built-in types; e.g. "i" for int, "f" for float.  For other types, the
previously mentioned name encoding should be used.

For templated types, the template parameters are encoded immediately after the
name.  If it is namespaced, it should be within the 'N' 'E' marker.  Template
parameters are encoded in "I<params>E", where each parameter is encoded using
the mentioned name encoding scheme.  Template parameters can contain literal
values like the '1' in the array type shown earlier.  There is special encoding
scheme for them to avoid leading digits.
"""

import re

from numba.cuda import types


# According the scheme, valid characters for mangled names are [a-zA-Z0-9_].
# We borrow the '_' as the escape character to encode invalid char into
# '_xx' where 'xx' is the hex codepoint.
_re_invalid_char = re.compile(r"[^a-z0-9_]", re.I)

PREFIX = "_Z"

# Numba types to mangled type code. These correspond with the codes listed in
# https://itanium-cxx-abi.github.io/cxx-abi/abi.html#mangling-builtin
N2CODE = {
    types.void: "v",
    types.boolean: "b",
    types.uint8: "h",
    types.int8: "a",
    types.uint16: "t",
    types.int16: "s",
    types.uint32: "j",
    types.int32: "i",
    types.uint64: "y",
    types.int64: "x",
    types.float16: "Dh",
    types.float32: "f",
    types.float64: "d",
}


def escape_string(text):
    """Escape the given string so that it only contains ASCII characters
    valid in mangled names: ``[a-zA-Z0-9_]``.

    Invalid characters are hex-escaped as ``_xx`` where *xx* is the two-digit
    hex code point.  Multibyte (non-ASCII) characters are first encoded to
    UTF-8 and each byte is escaped individually.

    Examples::

        >>> escape_string("hello")
        'hello'
        >>> escape_string("<lambda>")
        '_3clambda_3e'
    """

    def repl(m):
        return "".join(("_%02x" % ch) for ch in m.group(0).encode("utf8"))

    ret = re.sub(_re_invalid_char, repl, text)
    if not isinstance(ret, str):
        return ret.encode("ascii")
    return ret


_escape_string = escape_string


def _fix_lead_digit(text):
    """
    Fix text with leading digit
    """
    if text and text[0].isdigit():
        return "_" + text
    else:
        return text


def _len_encoded(string):
    """
    Prefix string with digit indicating the length.
    Add underscore if string is prefixed with digits.
    """
    string = _fix_lead_digit(string)
    return "%u%s" % (len(string), string)


def mangle_abi_tag(abi_tag: str) -> str:
    return "B" + _len_encoded(_escape_string(abi_tag))


def mangle_identifier(ident, template_params="", *, abi_tags=(), uid=None):
    """
    Mangle the identifier with optional template parameters and abi_tags.

    Note:

    This treats '.' as '::' in C++.
    """
    if uid is not None:
        # Add uid to abi-tags
        abi_tags = (f"v{uid}", *abi_tags)
    parts = [_len_encoded(_escape_string(x)) for x in ident.split(".")]
    enc_abi_tags = list(map(mangle_abi_tag, abi_tags))
    extras = template_params + "".join(enc_abi_tags)
    if len(parts) > 1:
        return "N%s%sE" % ("".join(parts), extras)
    else:
        return "%s%s" % (parts[0], extras)


def mangle_type_or_value(typ):
    """
    Mangle type parameter and arbitrary value.
    """
    # Handle numba types
    if isinstance(typ, types.Type):
        if typ in N2CODE:
            return N2CODE[typ]
        else:
            return mangle_templated_ident(*typ.mangling_args)
    # Handle integer literal
    elif isinstance(typ, int):
        return "Li%dE" % typ
    # Handle str as identifier
    elif isinstance(typ, str):
        return mangle_identifier(typ)
    # Otherwise
    else:
        enc = _escape_string(str(typ))
        return _len_encoded(enc)


# Alias
mangle_type = mangle_type_or_value
mangle_value = mangle_type_or_value


def mangle_templated_ident(identifier, parameters):
    """
    Mangle templated identifier.
    """
    template_params = (
        "I%sE" % "".join(map(mangle_type_or_value, parameters))
        if parameters
        else ""
    )
    return mangle_identifier(identifier, template_params)


def mangle_args(argtys):
    """
    Mangle sequence of Numba type objects and arbitrary values.
    """
    return "".join([mangle_type_or_value(t) for t in argtys])


def mangle(ident, argtys, *, abi_tags=(), uid=None):
    """
    Mangle identifier with Numba type objects and abi-tags.
    """
    return "".join(
        [
            PREFIX,
            mangle_identifier(ident, abi_tags=abi_tags, uid=uid),
            mangle_args(argtys),
        ]
    )


def prepend_namespace(mangled, ns):
    """
    Prepend namespace to mangled name.
    """
    if not mangled.startswith(PREFIX):
        raise ValueError("input is not a mangled name")
    elif mangled.startswith(PREFIX + "N"):
        # nested
        remaining = mangled[3:]
        ret = PREFIX + "N" + mangle_identifier(ns) + remaining
    else:
        # non-nested
        remaining = mangled[2:]
        head, tail = _split_mangled_ident(remaining)
        ret = PREFIX + "N" + mangle_identifier(ns) + head + "E" + tail
    return ret


def _split_mangled_ident(mangled):
    """
    Returns `(head, tail)` where `head` is the `<len> + <name>` encoded
    identifier and `tail` is the remaining.
    """
    ct = int(mangled)
    ctlen = len(str(ct))
    at = ctlen + ct
    return mangled[:at], mangled[at:]



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause

from numba.cuda import types
from numba.cuda import cgutils
from numba.cuda import itanium_mangler
from numba.cuda.core import imputils
from collections import namedtuple

from llvmlite import ir

int32_t = ir.IntType(32)
int64_t = ir.IntType(64)
errcode_t = int32_t


Status = namedtuple(
    "Status",
    (
        "code",
        # If the function returned ok (a value or None)
        "is_ok",
        # If the function returned None
        "is_none",
        # If the function errored out (== not is_ok)
        "is_error",
        # If the generator exited with StopIteration
        "is_stop_iteration",
        # If the function errored with an already set exception
        "is_python_exc",
        # If the function errored with a user exception
        "is_user_exc",
        # The pointer to the exception info structure (for user
        # exceptions)
        "excinfoptr",
    ),
)


def _const_int(code):
    return ir.Constant(errcode_t, code)


RETCODE_OK = _const_int(0)
RETCODE_EXC = _const_int(-1)
RETCODE_NONE = _const_int(-2)
# StopIteration
RETCODE_STOPIT = _const_int(-3)

FIRST_USEREXC = 1

RETCODE_USEREXC = _const_int(FIRST_USEREXC)


class BaseCallConv:
    needs_env = True

    def __init__(self, context):
        self.context = context

    def return_optional_value(self, builder, retty, valty, value):
        if valty == types.none:
            # Value is none
            self.return_native_none(builder)

        elif retty == valty:
            # Value is an optional, need a runtime switch
            optval = self.context.make_helper(builder, retty, value=value)

            validbit = cgutils.as_bool_bit(builder, optval.valid)
            with builder.if_then(validbit):
                retval = self.context.get_return_value(
                    builder, retty.type, optval.data
                )
                self.return_value(builder, retval)

            self.return_native_none(builder)

        elif not isinstance(valty, types.Optional):
            # Value is not an optional, need a cast
            if valty != retty.type:
                value = self.context.cast(
                    builder, value, fromty=valty, toty=retty.type
                )
            retval = self.context.get_return_value(builder, retty.type, value)
            self.return_value(builder, retval)

        else:
            raise NotImplementedError(
                "returning {0} for {1}".format(valty, retty)
            )

    def return_native_none(self, builder):
        self._return_errcode_raw(builder, RETCODE_NONE)

    def return_exc(self, builder):
        self._return_errcode_raw(builder, RETCODE_EXC)

    def return_stop_iteration(self, builder):
        self._return_errcode_raw(builder, RETCODE_STOPIT)

    def get_return_type(self, ty):
        """
        Get the actual type of the return argument for Numba type *ty*.
        """
        restype = self.context.data_model_manager[ty].get_return_type()
        return restype.as_pointer()

    def init_call_helper(self, builder):
        """
        Initialize and return a call helper object for the given builder.
        """
        ch = self._make_call_helper(builder)
        builder.__call_helper = ch
        return ch

    def _get_call_helper(self, builder):
        return builder.__call_helper

    def unpack_exception(self, builder, pyapi, status):
        return pyapi.unserialize(status.excinfoptr)

    def raise_error(self, builder, pyapi, status):
        """
        Given a non-ok *status*, raise the corresponding Python exception.
        """
        bbend = builder.function.append_basic_block()

        with builder.if_then(status.is_user_exc):
            # Unserialize user exception.
            # Make sure another error may not interfere.
            pyapi.err_clear()
            exc = self.unpack_exception(builder, pyapi, status)
            with cgutils.if_likely(builder, cgutils.is_not_null(builder, exc)):
                pyapi.raise_object(exc)  # steals ref
            builder.branch(bbend)

        with builder.if_then(status.is_stop_iteration):
            pyapi.err_set_none("PyExc_StopIteration")
            builder.branch(bbend)

        with builder.if_then(status.is_python_exc):
            # Error already raised => nothing to do
            builder.branch(bbend)

        pyapi.err_set_string(
            "PyExc_SystemError", "unknown error when calling native function"
        )
        builder.branch(bbend)

        builder.position_at_end(bbend)

    def decode_arguments(self, builder, argtypes, func):
        """
        Get the decoded (unpacked) Python arguments with *argtypes*
        from LLVM function *func*.  A tuple of LLVM values is returned.
        """
        raw_args = self.get_arguments(func)
        arginfo = self._get_arg_packer(argtypes)
        return arginfo.from_arguments(builder, raw_args)

    def _get_arg_packer(self, argtypes):
        """
        Get an argument packer for the given argument types.
        """
        return self.context.get_arg_packer(argtypes)

    def mangler(self, name, argtypes, *, abi_tags=(), uid=None):
        return itanium_mangler.mangle(
            name, argtypes, abi_tags=abi_tags, uid=uid
        )

    def call_internal_no_propagate(self, builder, fndesc, sig, args):
        """Similar to `.call_internal()` but does not handle or propagate
        the return status automatically.
        """
        llvm_mod = builder.module
        fn = fndesc.declare_function(llvm_mod)
        # Marshal the call using the callee's ABI.
        status, res = fndesc.call_conv.call_function(
            builder, fn, sig.return_type, sig.args, args
        )
        return status, res


class MinimalCallConv(BaseCallConv):
    """
    A minimal calling convention, suitable for e.g. GPU targets.
    The implemented function signature is:

        retcode_t (<Python return type>*, ... <Python arguments>)

    The return code will be one of the RETCODE_* constants or a
    function-specific user exception id (>= RETCODE_USEREXC).

    Caller is responsible for allocating a slot for the return value
    (passed as a pointer in the first argument).
    """

    def _make_call_helper(self, builder):
        return _MinimalCallHelper()

    def return_value(self, builder, retval):
        retptr = builder.function.args[0]
        assert retval.type == retptr.type.pointee, (
            str(retval.type),
            str(retptr.type.pointee),
        )
        builder.store(retval, retptr)
        self._return_errcode_raw(builder, RETCODE_OK)

    def return_user_exc(
        self, builder, exc, exc_args=None, loc=None, func_name=None
    ):
        if exc is not None and not issubclass(exc, BaseException):
            raise TypeError(
                "exc should be None or exception class, got %r" % (exc,)
            )
        if exc_args is not None and not isinstance(exc_args, tuple):
            raise TypeError(
                "exc_args should be None or tuple, got %r" % (exc_args,)
            )

        # Build excinfo struct
        if loc is not None:
            fname = loc._raw_function_name()
            if fname is None:
                # could be exec(<string>) or REPL, try func_name
                fname = func_name

            locinfo = (fname, loc.filename, loc.line)
            if None in locinfo:
                locinfo = None
        else:
            locinfo = None

        call_helper = self._get_call_helper(builder)
        exc_id = call_helper._add_exception(exc, exc_args, locinfo)
        self._return_errcode_raw(builder, _const_int(exc_id))

    def return_status_propagate(self, builder, status):
        self._return_errcode_raw(builder, status.code)

    def _return_errcode_raw(self, builder, code):
        if isinstance(code, int):
            code = _const_int(code)
        builder.ret(code)

    def _get_return_status(self, builder, code):
        """
        Given a return *code*, get a Status instance.
        """
        norm = builder.icmp_signed("==", code, RETCODE_OK)
        none = builder.icmp_signed("==", code, RETCODE_NONE)
        ok = builder.or_(norm, none)
        err = builder.not_(ok)
        exc = builder.icmp_signed("==", code, RETCODE_EXC)
        is_stop_iteration = builder.icmp_signed("==", code, RETCODE_STOPIT)
        is_user_exc = builder.icmp_signed(">=", code, RETCODE_USEREXC)

        status = Status(
            code=code,
            is_ok=ok,
            is_error=err,
            is_python_exc=exc,
            is_none=none,
            is_user_exc=is_user_exc,
            is_stop_iteration=is_stop_iteration,
            excinfoptr=None,
        )
        return status

    def get_function_type(self, restype, argtypes):
        """
        Get the implemented Function type for *restype* and *argtypes*.
        """
        arginfo = self._get_arg_packer(argtypes)
        argtypes = list(arginfo.argument_types)
        resptr = self.get_return_type(restype)
        fnty = ir.FunctionType(errcode_t, [resptr] + argtypes)
        return fnty

    def decorate_function(self, fn, args, fe_argtypes, noalias=False):
        """
        Set names and attributes of function arguments.
        """
        assert not noalias
        arginfo = self._get_arg_packer(fe_argtypes)
        arginfo.assign_names(self.get_arguments(fn), ["arg." + a for a in args])
        fn.args[0].name = ".ret"

    def get_arguments(self, func):
        """
        Get the Python-level arguments of LLVM *func*.
        """
        return func.args[1:]

    def call_function(self, builder, callee, resty, argtys, args):
        """
        Call the Numba-compiled *callee*.
        """
        retty = callee.args[0].type.pointee
        retvaltmp = cgutils.alloca_once(builder, retty)
        # initialize return value
        builder.store(cgutils.get_null_value(retty), retvaltmp)

        arginfo = self._get_arg_packer(argtys)
        args = arginfo.as_arguments(builder, args)
        realargs = [retvaltmp] + list(args)
        code = builder.call(callee, realargs)
        status = self._get_return_status(builder, code)
        retval = builder.load(retvaltmp)
        out = self.context.get_returned_value(builder, resty, retval)
        return status, out

    def call_internal(self, builder, fndesc, sig, args):
        """
        Given the function descriptor of an internally compiled function,
        emit a call to that function with the given arguments.
        """
        status, res = self.call_internal_no_propagate(
            builder, fndesc, sig, args
        )
        if status is not None:
            with cgutils.if_unlikely(builder, status.is_error):
                self.return_status_propagate(builder, status)

        res = imputils.fix_returning_optional(
            self.context, builder, sig, status, res
        )
        return res


class _MinimalCallHelper:
    """
    A call helper object for the "minimal" calling convention.
    User exceptions are represented as integer codes and stored in
    a mapping for retrieval from the caller.
    """

    def __init__(self):
        self.exceptions = {}

    def _add_exception(self, exc, exc_args, locinfo):
        """
        Add a new user exception to this helper. Returns an integer that can be
        used to refer to the added exception in future.

        Parameters
        ----------
        exc :
            exception type
        exc_args : None or tuple
            exception args
        locinfo : tuple
            location information
        """
        exc_id = len(self.exceptions) + FIRST_USEREXC
        self.exceptions[exc_id] = exc, exc_args, locinfo
        return exc_id

    def get_exception(self, exc_id):
        """
        Get information about a user exception. Returns a tuple of
        (exception type, exception args, location information).

        Parameters
        ----------
        id : integer
            The ID of the exception to look up
        """
        try:
            return self.exceptions[exc_id]
        except KeyError:
            msg = "unknown error %d in native function" % exc_id
            exc = SystemError
            exc_args = (msg,)
            locinfo = None
            return exc, exc_args, locinfo


class CUDACallConv(MinimalCallConv):
    def decorate_function(self, fn, args, fe_argtypes, noalias=False):
        """
        Set names and attributes of function arguments.
        """
        assert not noalias
        arginfo = self._get_arg_packer(fe_argtypes)
        # Do not prefix "arg." on argument name, so that nvvm compiler
        # can track debug info of argument more accurately
        arginfo.assign_names(self.get_arguments(fn), args)
        fn.args[0].name = ".ret"


class CUDACABICallConv(BaseCallConv):
    """
    Calling convention aimed at matching the CUDA C/C++ ABI. The implemented
    function signature is:

        <Python return type> (<Python arguments>)

    Exceptions are unsupported in this convention.

    No Numba environment pointer is needed because there is no error-status
    channel to propagate exceptions through.
    """

    needs_env = False

    def _make_call_helper(self, builder):
        # Call helpers are used to help report exceptions back to Python, so
        # none is required here.
        return None

    def return_value(self, builder, retval):
        expected_type = builder.function.ftype.return_type
        actual_type = retval.type

        # If types don't match, we need to cast
        if actual_type != expected_type:
            if isinstance(actual_type, ir.IntType) and isinstance(
                expected_type, ir.IntType
            ):
                if actual_type.width < expected_type.width:
                    # Zero-extend smaller integers to larger ones
                    retval = builder.zext(retval, expected_type)
                elif actual_type.width > expected_type.width:
                    # Truncate larger integers to smaller ones
                    retval = builder.trunc(retval, expected_type)

        return builder.ret(retval)

    def return_user_exc(
        self, builder, exc, exc_args=None, loc=None, func_name=None
    ):
        # C ABI has no status channel to propagate Python exceptions.
        return

    def return_status_propagate(self, builder, status):
        # C ABI has no status channel to propagate lower-frame failures.
        return

    def get_function_type(self, restype, argtypes):
        """
        Get the LLVM IR Function type for *restype* and *argtypes*.
        """
        arginfo = self._get_arg_packer(argtypes)
        argtypes = list(arginfo.argument_types)
        fnty = ir.FunctionType(self.get_return_type(restype), argtypes)
        return fnty

    def decorate_function(self, fn, args, fe_argtypes, noalias=False):
        """
        Set names and attributes of function arguments.
        """
        assert not noalias
        arginfo = self._get_arg_packer(fe_argtypes)
        arginfo.assign_names(self.get_arguments(fn), ["arg." + a for a in args])

    def get_arguments(self, func):
        """
        Get the Python-level arguments of LLVM *func*.
        """
        return func.args

    def call_function(self, builder, callee, resty, argtys, args):
        """
        Call the Numba-compiled *callee*.
        """
        arginfo = self._get_arg_packer(argtys)
        realargs = arginfo.as_arguments(builder, args)
        code = builder.call(callee, realargs)
        # No status required as we don't support exceptions or a distinct None
        # value in a C ABI.
        status = None
        out = self.context.get_returned_value(builder, resty, code)
        return status, out

    def call_internal(self, builder, fndesc, sig, args):
        """
        Given the function descriptor of an internally compiled function,
        emit a call to that function with the given arguments.
        """
        status, res = self.call_internal_no_propagate(
            builder, fndesc, sig, args
        )

        # CABI intentionally ignores lower-frame error codes.
        if not isinstance(sig.return_type, types.Optional):
            return res

        # A callee without a status channel cannot represent None.
        if status is None:
            return res

        # Flatten Optional[T] into plain T for CABI:
        # - if value is present, return it
        # - if value is None, return a default-initialized T
        value_type = sig.return_type.type
        default_value = self.context.get_constant_null(value_type)

        outptr = cgutils.alloca_once_value(builder, default_value)
        with builder.if_then(builder.not_(status.is_none)):
            builder.store(res, outptr)
        return builder.load(outptr)

    def get_return_type(self, ty):
        return self.context.data_model_manager[ty].get_return_type()

    def mangler(self, name, argtypes, *, abi_tags=None, uid=None):
        escaped = itanium_mangler.escape_string(name.split(".")[-1])
        if name.startswith(".NumbaEnv."):
            return f"_ZN08NumbaEnv{escaped}"
        return escaped


class ErrorModel:
    def __init__(self, call_conv):
        self.call_conv = call_conv

    def fp_zero_division(self, builder, exc_args=None, loc=None):
        if self.raise_on_fp_zero_division:
            self.call_conv.return_user_exc(
                builder,
                ZeroDivisionError,
                exc_args=exc_args,
                loc=loc,
            )
            return True
        return False


class PythonErrorModel(ErrorModel):
    """
    The Python error model.  Any invalid FP input raises an exception.
    """

    raise_on_fp_zero_division = True


class NumpyErrorModel(ErrorModel):
    """
    In the Numpy error model, floating-point errors don't raise an
    exception.  The FPU exception state is inspected by Numpy at the
    end of a ufunc's execution and a warning is raised if appropriate.

    Note there's no easy way to set the FPU exception state from LLVM.
    Instructions known to set an FP exception can be optimized away:
        https://llvm.org/bugs/show_bug.cgi?id=6050
        http://lists.llvm.org/pipermail/llvm-dev/2014-September/076918.html
        http://lists.llvm.org/pipermail/llvm-commits/Week-of-Mon-20140929/237997.html
    """

    raise_on_fp_zero_division = False


error_models = {
    "python": PythonErrorModel,
    "numpy": NumpyErrorModel,
}


def create_error_model(model_name, call_conv):
    """
    Create an error model instance for the given target context.
    """
    return error_models[model_name](call_conv)



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause

"""This file contains information on how to translate different ufuncs
into numba. It is a database of different ufuncs and how each of its
loops maps to a function that implements the inner kernel of that ufunc
(the inner kernel being the per-element function).

Use the function get_ufunc_info to get the information related to the
ufunc
"""

import numpy as np
import sys

# this is lazily initialized to avoid circular imports
IS_WIN32 = sys.platform.startswith("win32")
numpy_version = tuple(map(int, np.__version__.split(".")[:2]))
_ufunc_db = None


def _lazy_init_db():
    global _ufunc_db

    if _ufunc_db is None:
        _ufunc_db = {}
        _fill_ufunc_db(_ufunc_db)


def get_ufuncs():
    """obtain a list of supported ufuncs in the db"""
    _lazy_init_db()
    return _ufunc_db.keys()


def get_ufunc_info(ufunc_key):
    """get the lowering information for the ufunc with key ufunc_key.

    The lowering information is a dictionary that maps from a numpy
    loop string (as given by the ufunc types attribute) to a function
    that handles code generation for a scalar version of the ufunc
    (that is, generates the "per element" operation").

    raises a KeyError if the ufunc is not in the ufunc_db
    """
    _lazy_init_db()
    return _ufunc_db[ufunc_key]


def _fill_ufunc_db(ufunc_db):
    # some of these imports would cause a problem of circular
    # imports if done at global scope when importing the numba
    # module.
    from numba.cuda.np import npyfuncs
    from numba.cuda.np.math import cmathimpl, mathimpl, numbers
    from numba.cuda.np.numpy_support import numpy_version

    ufunc_db[np.isnat] = {
        # datetime & timedelta
        "M->?": npyfuncs.np_datetime_isnat_impl,
        "m->?": npyfuncs.np_datetime_isnat_impl,
    }

    ufunc_db[np.negative] = {
        "?->?": numbers.int_invert_impl,
        "b->b": numbers.int_negate_impl,
        "B->B": numbers.int_negate_impl,
        "h->h": numbers.int_negate_impl,
        "H->H": numbers.int_negate_impl,
        "i->i": numbers.int_negate_impl,
        "I->I": numbers.int_negate_impl,
        "l->l": numbers.int_negate_impl,
        "L->L": numbers.int_negate_impl,
        "q->q": numbers.int_negate_impl,
        "Q->Q": numbers.int_negate_impl,
        "f->f": numbers.real_negate_impl,
        "d->d": numbers.real_negate_impl,
        "F->F": numbers.complex_negate_impl,
        "D->D": numbers.complex_negate_impl,
    }

    ufunc_db[np.positive] = {
        "?->?": numbers.int_positive_impl,
        "b->b": numbers.int_positive_impl,
        "B->B": numbers.int_positive_impl,
        "h->h": numbers.int_positive_impl,
        "H->H": numbers.int_positive_impl,
        "i->i": numbers.int_positive_impl,
        "I->I": numbers.int_positive_impl,
        "l->l": numbers.int_positive_impl,
        "L->L": numbers.int_positive_impl,
        "q->q": numbers.int_positive_impl,
        "Q->Q": numbers.int_positive_impl,
        "f->f": numbers.real_positive_impl,
        "d->d": numbers.real_positive_impl,
        "F->F": numbers.complex_positive_impl,
        "D->D": numbers.complex_positive_impl,
    }

    ufunc_db[np.absolute] = {
        "?->?": numbers.int_abs_impl,
        "b->b": numbers.int_abs_impl,
        "B->B": numbers.uint_abs_impl,
        "h->h": numbers.int_abs_impl,
        "H->H": numbers.uint_abs_impl,
        "i->i": numbers.int_abs_impl,
        "I->I": numbers.uint_abs_impl,
        "l->l": numbers.int_abs_impl,
        "L->L": numbers.uint_abs_impl,
        "q->q": numbers.int_abs_impl,
        "Q->Q": numbers.uint_abs_impl,
        "f->f": numbers.real_abs_impl,
        "d->d": numbers.real_abs_impl,
        "F->f": numbers.complex_abs_impl,
        "D->d": numbers.complex_abs_impl,
    }

    ufunc_db[np.sign] = {
        "b->b": numbers.int_sign_impl,
        "B->B": numbers.int_sign_impl,
        "h->h": numbers.int_sign_impl,
        "H->H": numbers.int_sign_impl,
        "i->i": numbers.int_sign_impl,
        "I->I": numbers.int_sign_impl,
        "l->l": numbers.int_sign_impl,
        "L->L": numbers.int_sign_impl,
        "q->q": numbers.int_sign_impl,
        "Q->Q": numbers.int_sign_impl,
        "f->f": numbers.real_sign_impl,
        "d->d": numbers.real_sign_impl,
        "F->F": npyfuncs.np_complex_sign_impl,
        "D->D": npyfuncs.np_complex_sign_impl,
    }

    ufunc_db[np.add] = {
        "??->?": numbers.int_or_impl,
        "bb->b": numbers.int_add_impl,
        "BB->B": numbers.int_add_impl,
        "hh->h": numbers.int_add_impl,
        "HH->H": numbers.int_add_impl,
        "ii->i": numbers.int_add_impl,
        "II->I": numbers.int_add_impl,
        "ll->l": numbers.int_add_impl,
        "LL->L": numbers.int_add_impl,
        "qq->q": numbers.int_add_impl,
        "QQ->Q": numbers.int_add_impl,
        "ff->f": numbers.real_add_impl,
        "dd->d": numbers.real_add_impl,
        "FF->F": numbers.complex_add_impl,
        "DD->D": numbers.complex_add_impl,
    }

    ufunc_db[np.subtract] = {
        "??->?": numbers.int_xor_impl,
        "bb->b": numbers.int_sub_impl,
        "BB->B": numbers.int_sub_impl,
        "hh->h": numbers.int_sub_impl,
        "HH->H": numbers.int_sub_impl,
        "ii->i": numbers.int_sub_impl,
        "II->I": numbers.int_sub_impl,
        "ll->l": numbers.int_sub_impl,
        "LL->L": numbers.int_sub_impl,
        "qq->q": numbers.int_sub_impl,
        "QQ->Q": numbers.int_sub_impl,
        "ff->f": numbers.real_sub_impl,
        "dd->d": numbers.real_sub_impl,
        "FF->F": numbers.complex_sub_impl,
        "DD->D": numbers.complex_sub_impl,
    }

    ufunc_db[np.multiply] = {
        "??->?": numbers.int_and_impl,
        "bb->b": numbers.int_mul_impl,
        "BB->B": numbers.int_mul_impl,
        "hh->h": numbers.int_mul_impl,
        "HH->H": numbers.int_mul_impl,
        "ii->i": numbers.int_mul_impl,
        "II->I": numbers.int_mul_impl,
        "ll->l": numbers.int_mul_impl,
        "LL->L": numbers.int_mul_impl,
        "qq->q": numbers.int_mul_impl,
        "QQ->Q": numbers.int_mul_impl,
        "ff->f": numbers.real_mul_impl,
        "dd->d": numbers.real_mul_impl,
        "FF->F": numbers.complex_mul_impl,
        "DD->D": numbers.complex_mul_impl,
    }

    if np.divide != np.true_divide:
        ufunc_db[np.divide] = {
            "bb->b": npyfuncs.np_int_sdiv_impl,
            "BB->B": npyfuncs.np_int_udiv_impl,
            "hh->h": npyfuncs.np_int_sdiv_impl,
            "HH->H": npyfuncs.np_int_udiv_impl,
            "ii->i": npyfuncs.np_int_sdiv_impl,
            "II->I": npyfuncs.np_int_udiv_impl,
            "ll->l": npyfuncs.np_int_sdiv_impl,
            "LL->L": npyfuncs.np_int_udiv_impl,
            "qq->q": npyfuncs.np_int_sdiv_impl,
            "QQ->Q": npyfuncs.np_int_udiv_impl,
            "ff->f": npyfuncs.np_real_div_impl,
            "dd->d": npyfuncs.np_real_div_impl,
            "FF->F": npyfuncs.np_complex_div_impl,
            "DD->D": npyfuncs.np_complex_div_impl,
        }

    ufunc_db[np.true_divide] = {
        "bb->d": npyfuncs.np_int_truediv_impl,
        "BB->d": npyfuncs.np_int_truediv_impl,
        "hh->d": npyfuncs.np_int_truediv_impl,
        "HH->d": npyfuncs.np_int_truediv_impl,
        "ii->d": npyfuncs.np_int_truediv_impl,
        "II->d": npyfuncs.np_int_truediv_impl,
        "ll->d": npyfuncs.np_int_truediv_impl,
        "LL->d": npyfuncs.np_int_truediv_impl,
        "qq->d": npyfuncs.np_int_truediv_impl,
        "QQ->d": npyfuncs.np_int_truediv_impl,
        "ff->f": npyfuncs.np_real_div_impl,
        "dd->d": npyfuncs.np_real_div_impl,
        "FF->F": npyfuncs.np_complex_div_impl,
        "DD->D": npyfuncs.np_complex_div_impl,
    }

    ufunc_db[np.floor_divide] = {
        "bb->b": npyfuncs.np_int_sdiv_impl,
        "BB->B": npyfuncs.np_int_udiv_impl,
        "hh->h": npyfuncs.np_int_sdiv_impl,
        "HH->H": npyfuncs.np_int_udiv_impl,
        "ii->i": npyfuncs.np_int_sdiv_impl,
        "II->I": npyfuncs.np_int_udiv_impl,
        "ll->l": npyfuncs.np_int_sdiv_impl,
        "LL->L": npyfuncs.np_int_udiv_impl,
        "qq->q": npyfuncs.np_int_sdiv_impl,
        "QQ->Q": npyfuncs.np_int_udiv_impl,
        "ff->f": npyfuncs.np_real_floor_div_impl,
        "dd->d": npyfuncs.np_real_floor_div_impl,
    }

    ufunc_db[np.remainder] = {
        "bb->b": npyfuncs.np_int_srem_impl,
        "BB->B": npyfuncs.np_int_urem_impl,
        "hh->h": npyfuncs.np_int_srem_impl,
        "HH->H": npyfuncs.np_int_urem_impl,
        "ii->i": npyfuncs.np_int_srem_impl,
        "II->I": npyfuncs.np_int_urem_impl,
        "ll->l": npyfuncs.np_int_srem_impl,
        "LL->L": npyfuncs.np_int_urem_impl,
        "qq->q": npyfuncs.np_int_srem_impl,
        "QQ->Q": npyfuncs.np_int_urem_impl,
        "ff->f": npyfuncs.np_real_mod_impl,
        "dd->d": npyfuncs.np_real_mod_impl,
    }

    ufunc_db[np.divmod] = {
        "bb->bb": npyfuncs.np_int_sdivrem_impl,
        "BB->BB": npyfuncs.np_int_udivrem_impl,
        "hh->hh": npyfuncs.np_int_sdivrem_impl,
        "HH->HH": npyfuncs.np_int_udivrem_impl,
        "ii->ii": npyfuncs.np_int_sdivrem_impl,
        "II->II": npyfuncs.np_int_udivrem_impl,
        "ll->ll": npyfuncs.np_int_sdivrem_impl,
        "LL->LL": npyfuncs.np_int_udivrem_impl,
        "qq->qq": npyfuncs.np_int_sdivrem_impl,
        "QQ->QQ": npyfuncs.np_int_udivrem_impl,
        "ff->ff": npyfuncs.np_real_divmod_impl,
        "dd->dd": npyfuncs.np_real_divmod_impl,
    }

    ufunc_db[np.fmod] = {
        "bb->b": npyfuncs.np_int_fmod_impl,
        "BB->B": npyfuncs.np_int_fmod_impl,
        "hh->h": npyfuncs.np_int_fmod_impl,
        "HH->H": npyfuncs.np_int_fmod_impl,
        "ii->i": npyfuncs.np_int_fmod_impl,
        "II->I": npyfuncs.np_int_fmod_impl,
        "ll->l": npyfuncs.np_int_fmod_impl,
        "LL->L": npyfuncs.np_int_fmod_impl,
        "qq->q": npyfuncs.np_int_fmod_impl,
        "QQ->Q": npyfuncs.np_int_fmod_impl,
        "ff->f": npyfuncs.np_real_fmod_impl,
        "dd->d": npyfuncs.np_real_fmod_impl,
    }

    ufunc_db[np.logaddexp] = {
        "ff->f": npyfuncs.np_real_logaddexp_impl,
        "dd->d": npyfuncs.np_real_logaddexp_impl,
    }

    ufunc_db[np.logaddexp2] = {
        "ff->f": npyfuncs.np_real_logaddexp2_impl,
        "dd->d": npyfuncs.np_real_logaddexp2_impl,
    }

    ufunc_db[np.power] = {
        "bb->b": numbers.int_power_impl,
        "BB->B": numbers.int_power_impl,
        "hh->h": numbers.int_power_impl,
        "HH->H": numbers.int_power_impl,
        "ii->i": numbers.int_power_impl,
        "II->I": numbers.int_power_impl,
        "ll->l": numbers.int_power_impl,
        "LL->L": numbers.int_power_impl,
        "qq->q": numbers.int_power_impl,
        "QQ->Q": numbers.int_power_impl,
        # XXX we would like to use `int_power_impl` for real ** integer
        # as well (for better performance), but the current ufunc typing
        # rules forbid that
        "ff->f": numbers.real_power_impl,
        "dd->d": numbers.real_power_impl,
        "FF->F": npyfuncs.np_complex_power_impl,
        "DD->D": npyfuncs.np_complex_power_impl,
    }

    ufunc_db[np.float_power] = {
        "ff->f": npyfuncs.real_float_power_impl,
        "dd->d": npyfuncs.real_float_power_impl,
        "FF->F": npyfuncs.np_complex_float_power_impl,
        "DD->D": npyfuncs.np_complex_float_power_impl,
    }

    ufunc_db[np.gcd] = {
        "bb->b": npyfuncs.np_gcd_impl,
        "BB->B": npyfuncs.np_gcd_impl,
        "hh->h": npyfuncs.np_gcd_impl,
        "HH->H": npyfuncs.np_gcd_impl,
        "ii->i": npyfuncs.np_gcd_impl,
        "II->I": npyfuncs.np_gcd_impl,
        "ll->l": npyfuncs.np_gcd_impl,
        "LL->L": npyfuncs.np_gcd_impl,
        "qq->q": npyfuncs.np_gcd_impl,
        "QQ->Q": npyfuncs.np_gcd_impl,
    }

    ufunc_db[np.lcm] = {
        "bb->b": npyfuncs.np_lcm_impl,
        "BB->B": npyfuncs.np_lcm_impl,
        "hh->h": npyfuncs.np_lcm_impl,
        "HH->H": npyfuncs.np_lcm_impl,
        "ii->i": npyfuncs.np_lcm_impl,
        "II->I": npyfuncs.np_lcm_impl,
        "ll->l": npyfuncs.np_lcm_impl,
        "LL->L": npyfuncs.np_lcm_impl,
        "qq->q": npyfuncs.np_lcm_impl,
        "QQ->Q": npyfuncs.np_lcm_impl,
    }

    ufunc_db[np.rint] = {
        "f->f": npyfuncs.np_real_rint_impl,
        "d->d": npyfuncs.np_real_rint_impl,
        "F->F": npyfuncs.np_complex_rint_impl,
        "D->D": npyfuncs.np_complex_rint_impl,
    }

    ufunc_db[np.conjugate] = {
        "b->b": numbers.real_conjugate_impl,
        "B->B": numbers.real_conjugate_impl,
        "h->h": numbers.real_conjugate_impl,
        "H->H": numbers.real_conjugate_impl,
        "i->i": numbers.real_conjugate_impl,
        "I->I": numbers.real_conjugate_impl,
        "l->l": numbers.real_conjugate_impl,
        "L->L": numbers.real_conjugate_impl,
        "q->q": numbers.real_conjugate_impl,
        "Q->Q": numbers.real_conjugate_impl,
        "f->f": numbers.real_conjugate_impl,
        "d->d": numbers.real_conjugate_impl,
        "F->F": numbers.complex_conjugate_impl,
        "D->D": numbers.complex_conjugate_impl,
    }

    ufunc_db[np.exp] = {
        "f->f": npyfuncs.np_real_exp_impl,
        "d->d": npyfuncs.np_real_exp_impl,
        "F->F": npyfuncs.np_complex_exp_impl,
        "D->D": npyfuncs.np_complex_exp_impl,
    }

    ufunc_db[np.exp2] = {
        "f->f": npyfuncs.np_real_exp2_impl,
        "d->d": npyfuncs.np_real_exp2_impl,
        "F->F": npyfuncs.np_complex_exp2_impl,
        "D->D": npyfuncs.np_complex_exp2_impl,
    }

    ufunc_db[np.log] = {
        "f->f": npyfuncs.np_real_log_impl,
        "d->d": npyfuncs.np_real_log_impl,
        "F->F": npyfuncs.np_complex_log_impl,
        "D->D": npyfuncs.np_complex_log_impl,
    }

    ufunc_db[np.log2] = {
        "f->f": npyfuncs.np_real_log2_impl,
        "d->d": npyfuncs.np_real_log2_impl,
        "F->F": npyfuncs.np_complex_log2_impl,
        "D->D": npyfuncs.np_complex_log2_impl,
    }

    ufunc_db[np.log10] = {
        "f->f": npyfuncs.np_real_log10_impl,
        "d->d": npyfuncs.np_real_log10_impl,
        "F->F": npyfuncs.np_complex_log10_impl,
        "D->D": npyfuncs.np_complex_log10_impl,
    }

    ufunc_db[np.expm1] = {
        "f->f": npyfuncs.np_real_expm1_impl,
        "d->d": npyfuncs.np_real_expm1_impl,
        "F->F": npyfuncs.np_complex_expm1_impl,
        "D->D": npyfuncs.np_complex_expm1_impl,
    }

    ufunc_db[np.log1p] = {
        "f->f": npyfuncs.np_real_log1p_impl,
        "d->d": npyfuncs.np_real_log1p_impl,
        "F->F": npyfuncs.np_complex_log1p_impl,
        "D->D": npyfuncs.np_complex_log1p_impl,
    }

    ufunc_db[np.sqrt] = {
        "f->f": npyfuncs.np_real_sqrt_impl,
        "d->d": npyfuncs.np_real_sqrt_impl,
        "F->F": npyfuncs.np_complex_sqrt_impl,
        "D->D": npyfuncs.np_complex_sqrt_impl,
    }

    ufunc_db[np.square] = {
        "b->b": npyfuncs.np_int_square_impl,
        "B->B": npyfuncs.np_int_square_impl,
        "h->h": npyfuncs.np_int_square_impl,
        "H->H": npyfuncs.np_int_square_impl,
        "i->i": npyfuncs.np_int_square_impl,
        "I->I": npyfuncs.np_int_square_impl,
        "l->l": npyfuncs.np_int_square_impl,
        "L->L": npyfuncs.np_int_square_impl,
        "q->q": npyfuncs.np_int_square_impl,
        "Q->Q": npyfuncs.np_int_square_impl,
        "f->f": npyfuncs.np_real_square_impl,
        "d->d": npyfuncs.np_real_square_impl,
        "F->F": npyfuncs.np_complex_square_impl,
        "D->D": npyfuncs.np_complex_square_impl,
    }

    ufunc_db[np.cbrt] = {
        "f->f": npyfuncs.np_real_cbrt_impl,
        "d->d": npyfuncs.np_real_cbrt_impl,
    }

    ufunc_db[np.reciprocal] = {
        "b->b": npyfuncs.np_int_reciprocal_impl,
        "B->B": npyfuncs.np_int_reciprocal_impl,
        "h->h": npyfuncs.np_int_reciprocal_impl,
        "H->H": npyfuncs.np_int_reciprocal_impl,
        "i->i": npyfuncs.np_int_reciprocal_impl,
        "I->I": npyfuncs.np_int_reciprocal_impl,
        "l->l": npyfuncs.np_int_reciprocal_impl,
        "L->L": npyfuncs.np_int_reciprocal_impl,
        "q->q": npyfuncs.np_int_reciprocal_impl,
        "Q->Q": npyfuncs.np_int_reciprocal_impl,
        "f->f": npyfuncs.np_real_reciprocal_impl,
        "d->d": npyfuncs.np_real_reciprocal_impl,
        "F->F": npyfuncs.np_complex_reciprocal_impl,
        "D->D": npyfuncs.np_complex_reciprocal_impl,
    }

    ufunc_db[np.sin] = {
        "f->f": npyfuncs.np_real_sin_impl,
        "d->d": npyfuncs.np_real_sin_impl,
        "F->F": npyfuncs.np_complex_sin_impl,
        "D->D": npyfuncs.np_complex_sin_impl,
    }

    ufunc_db[np.cos] = {
        "f->f": npyfuncs.np_real_cos_impl,
        "d->d": npyfuncs.np_real_cos_impl,
        "F->F": npyfuncs.np_complex_cos_impl,
        "D->D": npyfuncs.np_complex_cos_impl,
    }

    tan_impl = cmathimpl.tan_impl

    ufunc_db[np.tan] = {
        "f->f": npyfuncs.np_real_tan_impl,
        "d->d": npyfuncs.np_real_tan_impl,
        "F->F": tan_impl,
        "D->D": tan_impl,
    }

    arcsin_impl = cmathimpl.asin_impl

    ufunc_db[np.arcsin] = {
        "f->f": npyfuncs.np_real_asin_impl,
        "d->d": npyfuncs.np_real_asin_impl,
        "F->F": arcsin_impl,
        "D->D": arcsin_impl,
    }

    ufunc_db[np.arccos] = {
        "f->f": npyfuncs.np_real_acos_impl,
        "d->d": npyfuncs.np_real_acos_impl,
        "F->F": cmathimpl.acos_impl,
        "D->D": cmathimpl.acos_impl,
    }

    arctan_impl = cmathimpl.atan_impl

    ufunc_db[np.arctan] = {
        "f->f": npyfuncs.np_real_atan_impl,
        "d->d": npyfuncs.np_real_atan_impl,
        "F->F": arctan_impl,
        "D->D": arctan_impl,
    }

    ufunc_db[np.arctan2] = {
        "ff->f": npyfuncs.np_real_atan2_impl,
        "dd->d": npyfuncs.np_real_atan2_impl,
    }

    ufunc_db[np.hypot] = {
        "ff->f": npyfuncs.np_real_hypot_impl,
        "dd->d": npyfuncs.np_real_hypot_impl,
    }

    ufunc_db[np.sinh] = {
        "f->f": npyfuncs.np_real_sinh_impl,
        "d->d": npyfuncs.np_real_sinh_impl,
        "F->F": npyfuncs.np_complex_sinh_impl,
        "D->D": npyfuncs.np_complex_sinh_impl,
    }

    ufunc_db[np.cosh] = {
        "f->f": npyfuncs.np_real_cosh_impl,
        "d->d": npyfuncs.np_real_cosh_impl,
        "F->F": npyfuncs.np_complex_cosh_impl,
        "D->D": npyfuncs.np_complex_cosh_impl,
    }

    ufunc_db[np.tanh] = {
        "f->f": npyfuncs.np_real_tanh_impl,
        "d->d": npyfuncs.np_real_tanh_impl,
        "F->F": npyfuncs.np_complex_tanh_impl,
        "D->D": npyfuncs.np_complex_tanh_impl,
    }

    arcsinh_impl = cmathimpl.asinh_impl

    ufunc_db[np.arcsinh] = {
        "f->f": npyfuncs.np_real_asinh_impl,
        "d->d": npyfuncs.np_real_asinh_impl,
        "F->F": arcsinh_impl,
        "D->D": arcsinh_impl,
    }

    ufunc_db[np.arccosh] = {
        "f->f": npyfuncs.np_real_acosh_impl,
        "d->d": npyfuncs.np_real_acosh_impl,
        "F->F": npyfuncs.np_complex_acosh_impl,
        "D->D": npyfuncs.np_complex_acosh_impl,
    }

    arctanh_impl = cmathimpl.atanh_impl

    ufunc_db[np.arctanh] = {
        "f->f": npyfuncs.np_real_atanh_impl,
        "d->d": npyfuncs.np_real_atanh_impl,
        "F->F": arctanh_impl,
        "D->D": arctanh_impl,
    }

    ufunc_db[np.deg2rad] = {
        "f->f": mathimpl.radians_float_impl,
        "d->d": mathimpl.radians_float_impl,
    }

    ufunc_db[np.radians] = ufunc_db[np.deg2rad]

    ufunc_db[np.rad2deg] = {
        "f->f": mathimpl.degrees_float_impl,
        "d->d": mathimpl.degrees_float_impl,
    }

    ufunc_db[np.degrees] = ufunc_db[np.rad2deg]

    ufunc_db[np.floor] = {
        "f->f": npyfuncs.np_real_floor_impl,
        "d->d": npyfuncs.np_real_floor_impl,
    }
    if numpy_version >= (2, 1):
        ufunc_db[np.floor].update(
            {
                "?->?": numbers.identity_impl,
                "b->b": numbers.identity_impl,
                "B->B": numbers.identity_impl,
                "h->h": numbers.identity_impl,
                "H->H": numbers.identity_impl,
                "i->i": numbers.identity_impl,
                "I->I": numbers.identity_impl,
                "l->l": numbers.identity_impl,
                "L->L": numbers.identity_impl,
                "q->q": numbers.identity_impl,
                "Q->Q": numbers.identity_impl,
            }
        )

    ufunc_db[np.ceil] = {
        "f->f": npyfuncs.np_real_ceil_impl,
        "d->d": npyfuncs.np_real_ceil_impl,
    }
    if numpy_version >= (2, 1):
        ufunc_db[np.ceil].update(
            {
                "?->?": numbers.identity_impl,
                "b->b": numbers.identity_impl,
                "B->B": numbers.identity_impl,
                "h->h": numbers.identity_impl,
                "H->H": numbers.identity_impl,
                "i->i": numbers.identity_impl,
                "I->I": numbers.identity_impl,
                "l->l": numbers.identity_impl,
                "L->L": numbers.identity_impl,
                "q->q": numbers.identity_impl,
                "Q->Q": numbers.identity_impl,
            }
        )

    ufunc_db[np.trunc] = {
        "f->f": npyfuncs.np_real_trunc_impl,
        "d->d": npyfuncs.np_real_trunc_impl,
    }
    if numpy_version >= (2, 1):
        ufunc_db[np.trunc].update(
            {
                "?->?": numbers.identity_impl,
                "b->b": numbers.identity_impl,
                "B->B": numbers.identity_impl,
                "h->h": numbers.identity_impl,
                "H->H": numbers.identity_impl,
                "i->i": numbers.identity_impl,
                "I->I": numbers.identity_impl,
                "l->l": numbers.identity_impl,
                "L->L": numbers.identity_impl,
                "q->q": numbers.identity_impl,
                "Q->Q": numbers.identity_impl,
            }
        )

    ufunc_db[np.fabs] = {
        "f->f": npyfuncs.np_real_fabs_impl,
        "d->d": npyfuncs.np_real_fabs_impl,
    }

    # logical ufuncs
    ufunc_db[np.greater] = {
        "??->?": numbers.int_ugt_impl,
        "bb->?": numbers.int_sgt_impl,
        "BB->?": numbers.int_ugt_impl,
        "hh->?": numbers.int_sgt_impl,
        "HH->?": numbers.int_ugt_impl,
        "ii->?": numbers.int_sgt_impl,
        "II->?": numbers.int_ugt_impl,
        "ll->?": numbers.int_sgt_impl,
        "LL->?": numbers.int_ugt_impl,
        "qq->?": numbers.int_sgt_impl,
        "QQ->?": numbers.int_ugt_impl,
        "ff->?": numbers.real_gt_impl,
        "dd->?": numbers.real_gt_impl,
        "FF->?": npyfuncs.np_complex_gt_impl,
        "DD->?": npyfuncs.np_complex_gt_impl,
    }
    if numpy_version >= (1, 25):
        ufunc_db[np.greater].update(
            {
                "qQ->?": numbers.int_signed_unsigned_cmp(">"),
                "Qq->?": numbers.int_unsigned_signed_cmp(">"),
            }
        )

    ufunc_db[np.greater_equal] = {
        "??->?": numbers.int_uge_impl,
        "bb->?": numbers.int_sge_impl,
        "BB->?": numbers.int_uge_impl,
        "hh->?": numbers.int_sge_impl,
        "HH->?": numbers.int_uge_impl,
        "ii->?": numbers.int_sge_impl,
        "II->?": numbers.int_uge_impl,
        "ll->?": numbers.int_sge_impl,
        "LL->?": numbers.int_uge_impl,
        "qq->?": numbers.int_sge_impl,
        "QQ->?": numbers.int_uge_impl,
        "ff->?": numbers.real_ge_impl,
        "dd->?": numbers.real_ge_impl,
        "FF->?": npyfuncs.np_complex_ge_impl,
        "DD->?": npyfuncs.np_complex_ge_impl,
    }
    if numpy_version >= (1, 25):
        ufunc_db[np.greater_equal].update(
            {
                "qQ->?": numbers.int_signed_unsigned_cmp(">="),
                "Qq->?": numbers.int_unsigned_signed_cmp(">="),
            }
        )

    ufunc_db[np.less] = {
        "??->?": numbers.int_ult_impl,
        "bb->?": numbers.int_slt_impl,
        "BB->?": numbers.int_ult_impl,
        "hh->?": numbers.int_slt_impl,
        "HH->?": numbers.int_ult_impl,
        "ii->?": numbers.int_slt_impl,
        "II->?": numbers.int_ult_impl,
        "ll->?": numbers.int_slt_impl,
        "LL->?": numbers.int_ult_impl,
        "qq->?": numbers.int_slt_impl,
        "QQ->?": numbers.int_ult_impl,
        "ff->?": numbers.real_lt_impl,
        "dd->?": numbers.real_lt_impl,
        "FF->?": npyfuncs.np_complex_lt_impl,
        "DD->?": npyfuncs.np_complex_lt_impl,
    }
    if numpy_version >= (1, 25):
        ufunc_db[np.less].update(
            {
                "qQ->?": numbers.int_signed_unsigned_cmp("<"),
                "Qq->?": numbers.int_unsigned_signed_cmp("<"),
            }
        )

    ufunc_db[np.less_equal] = {
        "??->?": numbers.int_ule_impl,
        "bb->?": numbers.int_sle_impl,
        "BB->?": numbers.int_ule_impl,
        "hh->?": numbers.int_sle_impl,
        "HH->?": numbers.int_ule_impl,
        "ii->?": numbers.int_sle_impl,
        "II->?": numbers.int_ule_impl,
        "ll->?": numbers.int_sle_impl,
        "LL->?": numbers.int_ule_impl,
        "qq->?": numbers.int_sle_impl,
        "QQ->?": numbers.int_ule_impl,
        "ff->?": numbers.real_le_impl,
        "dd->?": numbers.real_le_impl,
        "FF->?": npyfuncs.np_complex_le_impl,
        "DD->?": npyfuncs.np_complex_le_impl,
    }
    if numpy_version >= (1, 25):
        ufunc_db[np.less_equal].update(
            {
                "qQ->?": numbers.int_signed_unsigned_cmp("<="),
                "Qq->?": numbers.int_unsigned_signed_cmp("<="),
            }
        )

    ufunc_db[np.not_equal] = {
        "??->?": numbers.int_ne_impl,
        "bb->?": numbers.int_ne_impl,
        "BB->?": numbers.int_ne_impl,
        "hh->?": numbers.int_ne_impl,
        "HH->?": numbers.int_ne_impl,
        "ii->?": numbers.int_ne_impl,
        "II->?": numbers.int_ne_impl,
        "ll->?": numbers.int_ne_impl,
        "LL->?": numbers.int_ne_impl,
        "qq->?": numbers.int_ne_impl,
        "QQ->?": numbers.int_ne_impl,
        "ff->?": numbers.real_ne_impl,
        "dd->?": numbers.real_ne_impl,
        "FF->?": npyfuncs.np_complex_ne_impl,
        "DD->?": npyfuncs.np_complex_ne_impl,
    }
    if numpy_version >= (1, 25):
        ufunc_db[np.not_equal].update(
            {
                "qQ->?": numbers.int_signed_unsigned_cmp("!="),
                "Qq->?": numbers.int_unsigned_signed_cmp("!="),
            }
        )

    ufunc_db[np.equal] = {
        "??->?": numbers.int_eq_impl,
        "bb->?": numbers.int_eq_impl,
        "BB->?": numbers.int_eq_impl,
        "hh->?": numbers.int_eq_impl,
        "HH->?": numbers.int_eq_impl,
        "ii->?": numbers.int_eq_impl,
        "II->?": numbers.int_eq_impl,
        "ll->?": numbers.int_eq_impl,
        "LL->?": numbers.int_eq_impl,
        "qq->?": numbers.int_eq_impl,
        "QQ->?": numbers.int_eq_impl,
        "ff->?": numbers.real_eq_impl,
        "dd->?": numbers.real_eq_impl,
        "FF->?": npyfuncs.np_complex_eq_impl,
        "DD->?": npyfuncs.np_complex_eq_impl,
    }
    if numpy_version >= (1, 25):
        ufunc_db[np.equal].update(
            {
                "qQ->?": numbers.int_signed_unsigned_cmp("=="),
                "Qq->?": numbers.int_unsigned_signed_cmp("=="),
            }
        )

    ufunc_db[np.logical_and] = {
        "??->?": npyfuncs.np_logical_and_impl,
        "bb->?": npyfuncs.np_logical_and_impl,
        "BB->?": npyfuncs.np_logical_and_impl,
        "hh->?": npyfuncs.np_logical_and_impl,
        "HH->?": npyfuncs.np_logical_and_impl,
        "ii->?": npyfuncs.np_logical_and_impl,
        "II->?": npyfuncs.np_logical_and_impl,
        "ll->?": npyfuncs.np_logical_and_impl,
        "LL->?": npyfuncs.np_logical_and_impl,
        "qq->?": npyfuncs.np_logical_and_impl,
        "QQ->?": npyfuncs.np_logical_and_impl,
        "ff->?": npyfuncs.np_logical_and_impl,
        "dd->?": npyfuncs.np_logical_and_impl,
        "FF->?": npyfuncs.np_complex_logical_and_impl,
        "DD->?": npyfuncs.np_complex_logical_and_impl,
    }

    ufunc_db[np.logical_or] = {
        "??->?": npyfuncs.np_logical_or_impl,
        "bb->?": npyfuncs.np_logical_or_impl,
        "BB->?": npyfuncs.np_logical_or_impl,
        "hh->?": npyfuncs.np_logical_or_impl,
        "HH->?": npyfuncs.np_logical_or_impl,
        "ii->?": npyfuncs.np_logical_or_impl,
        "II->?": npyfuncs.np_logical_or_impl,
        "ll->?": npyfuncs.np_logical_or_impl,
        "LL->?": npyfuncs.np_logical_or_impl,
        "qq->?": npyfuncs.np_logical_or_impl,
        "QQ->?": npyfuncs.np_logical_or_impl,
        "ff->?": npyfuncs.np_logical_or_impl,
        "dd->?": npyfuncs.np_logical_or_impl,
        "FF->?": npyfuncs.np_complex_logical_or_impl,
        "DD->?": npyfuncs.np_complex_logical_or_impl,
    }

    ufunc_db[np.logical_xor] = {
        "??->?": npyfuncs.np_logical_xor_impl,
        "bb->?": npyfuncs.np_logical_xor_impl,
        "BB->?": npyfuncs.np_logical_xor_impl,
        "hh->?": npyfuncs.np_logical_xor_impl,
        "HH->?": npyfuncs.np_logical_xor_impl,
        "ii->?": npyfuncs.np_logical_xor_impl,
        "II->?": npyfuncs.np_logical_xor_impl,
        "ll->?": npyfuncs.np_logical_xor_impl,
        "LL->?": npyfuncs.np_logical_xor_impl,
        "qq->?": npyfuncs.np_logical_xor_impl,
        "QQ->?": npyfuncs.np_logical_xor_impl,
        "ff->?": npyfuncs.np_logical_xor_impl,
        "dd->?": npyfuncs.np_logical_xor_impl,
        "FF->?": npyfuncs.np_complex_logical_xor_impl,
        "DD->?": npyfuncs.np_complex_logical_xor_impl,
    }

    ufunc_db[np.logical_not] = {
        "?->?": npyfuncs.np_logical_not_impl,
        "b->?": npyfuncs.np_logical_not_impl,
        "B->?": npyfuncs.np_logical_not_impl,
        "h->?": npyfuncs.np_logical_not_impl,
        "H->?": npyfuncs.np_logical_not_impl,
        "i->?": npyfuncs.np_logical_not_impl,
        "I->?": npyfuncs.np_logical_not_impl,
        "l->?": npyfuncs.np_logical_not_impl,
        "L->?": npyfuncs.np_logical_not_impl,
        "q->?": npyfuncs.np_logical_not_impl,
        "Q->?": npyfuncs.np_logical_not_impl,
        "f->?": npyfuncs.np_logical_not_impl,
        "d->?": npyfuncs.np_logical_not_impl,
        "F->?": npyfuncs.np_complex_logical_not_impl,
        "D->?": npyfuncs.np_complex_logical_not_impl,
    }

    ufunc_db[np.maximum] = {
        "??->?": npyfuncs.np_logical_or_impl,
        "bb->b": npyfuncs.np_int_smax_impl,
        "BB->B": npyfuncs.np_int_umax_impl,
        "hh->h": npyfuncs.np_int_smax_impl,
        "HH->H": npyfuncs.np_int_umax_impl,
        "ii->i": npyfuncs.np_int_smax_impl,
        "II->I": npyfuncs.np_int_umax_impl,
        "ll->l": npyfuncs.np_int_smax_impl,
        "LL->L": npyfuncs.np_int_umax_impl,
        "qq->q": npyfuncs.np_int_smax_impl,
        "QQ->Q": npyfuncs.np_int_umax_impl,
        "ff->f": npyfuncs.np_real_maximum_impl,
        "dd->d": npyfuncs.np_real_maximum_impl,
        "FF->F": npyfuncs.np_complex_maximum_impl,
        "DD->D": npyfuncs.np_complex_maximum_impl,
    }

    ufunc_db[np.minimum] = {
        "??->?": npyfuncs.np_logical_and_impl,
        "bb->b": npyfuncs.np_int_smin_impl,
        "BB->B": npyfuncs.np_int_umin_impl,
        "hh->h": npyfuncs.np_int_smin_impl,
        "HH->H": npyfuncs.np_int_umin_impl,
        "ii->i": npyfuncs.np_int_smin_impl,
        "II->I": npyfuncs.np_int_umin_impl,
        "ll->l": npyfuncs.np_int_smin_impl,
        "LL->L": npyfuncs.np_int_umin_impl,
        "qq->q": npyfuncs.np_int_smin_impl,
        "QQ->Q": npyfuncs.np_int_umin_impl,
        "ff->f": npyfuncs.np_real_minimum_impl,
        "dd->d": npyfuncs.np_real_minimum_impl,
        "FF->F": npyfuncs.np_complex_minimum_impl,
        "DD->D": npyfuncs.np_complex_minimum_impl,
    }

    ufunc_db[np.fmax] = {
        "??->?": npyfuncs.np_logical_or_impl,
        "bb->b": npyfuncs.np_int_smax_impl,
        "BB->B": npyfuncs.np_int_umax_impl,
        "hh->h": npyfuncs.np_int_smax_impl,
        "HH->H": npyfuncs.np_int_umax_impl,
        "ii->i": npyfuncs.np_int_smax_impl,
        "II->I": npyfuncs.np_int_umax_impl,
        "ll->l": npyfuncs.np_int_smax_impl,
        "LL->L": npyfuncs.np_int_umax_impl,
        "qq->q": npyfuncs.np_int_smax_impl,
        "QQ->Q": npyfuncs.np_int_umax_impl,
        "ff->f": npyfuncs.np_real_fmax_impl,
        "dd->d": npyfuncs.np_real_fmax_impl,
        "FF->F": npyfuncs.np_complex_fmax_impl,
        "DD->D": npyfuncs.np_complex_fmax_impl,
    }

    ufunc_db[np.fmin] = {
        "??->?": npyfuncs.np_logical_and_impl,
        "bb->b": npyfuncs.np_int_smin_impl,
        "BB->B": npyfuncs.np_int_umin_impl,
        "hh->h": npyfuncs.np_int_smin_impl,
        "HH->H": npyfuncs.np_int_umin_impl,
        "ii->i": npyfuncs.np_int_smin_impl,
        "II->I": npyfuncs.np_int_umin_impl,
        "ll->l": npyfuncs.np_int_smin_impl,
        "LL->L": npyfuncs.np_int_umin_impl,
        "qq->q": npyfuncs.np_int_smin_impl,
        "QQ->Q": npyfuncs.np_int_umin_impl,
        "ff->f": npyfuncs.np_real_fmin_impl,
        "dd->d": npyfuncs.np_real_fmin_impl,
        "FF->F": npyfuncs.np_complex_fmin_impl,
        "DD->D": npyfuncs.np_complex_fmin_impl,
    }

    # misc floating functions
    ufunc_db[np.isnan] = {
        "f->?": npyfuncs.np_real_isnan_impl,
        "d->?": npyfuncs.np_real_isnan_impl,
        "F->?": npyfuncs.np_complex_isnan_impl,
        "D->?": npyfuncs.np_complex_isnan_impl,
        # int8
        "b->?": npyfuncs.np_int_isnan_impl,
        "B->?": npyfuncs.np_int_isnan_impl,
        # int16
        "h->?": npyfuncs.np_int_isnan_impl,
        "H->?": npyfuncs.np_int_isnan_impl,
        # int32
        "i->?": npyfuncs.np_int_isnan_impl,
        "I->?": npyfuncs.np_int_isnan_impl,
        # int64
        "l->?": npyfuncs.np_int_isnan_impl,
        "L->?": npyfuncs.np_int_isnan_impl,
        # intp
        "q->?": npyfuncs.np_int_isnan_impl,
        "Q->?": npyfuncs.np_int_isnan_impl,
        # boolean
        "?->?": npyfuncs.np_int_isnan_impl,
        # datetime & timedelta
        "m->?": npyfuncs.np_datetime_isnat_impl,
        "M->?": npyfuncs.np_datetime_isnat_impl,
    }

    ufunc_db[np.isinf] = {
        "f->?": npyfuncs.np_real_isinf_impl,
        "d->?": npyfuncs.np_real_isinf_impl,
        "F->?": npyfuncs.np_complex_isinf_impl,
        "D->?": npyfuncs.np_complex_isinf_impl,
        # int8
        "b->?": npyfuncs.np_int_isinf_impl,
        "B->?": npyfuncs.np_int_isinf_impl,
        # int16
        "h->?": npyfuncs.np_int_isinf_impl,
        "H->?": npyfuncs.np_int_isinf_impl,
        # int32
        "i->?": npyfuncs.np_int_isinf_impl,
        "I->?": npyfuncs.np_int_isinf_impl,
        # int64
        "l->?": npyfuncs.np_int_isinf_impl,
        "L->?": npyfuncs.np_int_isinf_impl,
        # intp
        "q->?": npyfuncs.np_int_isinf_impl,
        "Q->?": npyfuncs.np_int_isinf_impl,
        # boolean
        "?->?": npyfuncs.np_int_isinf_impl,
        # datetime & timedelta
        "m->?": npyfuncs.np_int_isinf_impl,
        "M->?": npyfuncs.np_int_isinf_impl,
    }

    ufunc_db[np.isfinite] = {
        "f->?": npyfuncs.np_real_isfinite_impl,
        "d->?": npyfuncs.np_real_isfinite_impl,
        "F->?": npyfuncs.np_complex_isfinite_impl,
        "D->?": npyfuncs.np_complex_isfinite_impl,
        # int8
        "b->?": npyfuncs.np_int_isfinite_impl,
        "B->?": npyfuncs.np_int_isfinite_impl,
        # int16
        "h->?": npyfuncs.np_int_isfinite_impl,
        "H->?": npyfuncs.np_int_isfinite_impl,
        # int32
        "i->?": npyfuncs.np_int_isfinite_impl,
        "I->?": npyfuncs.np_int_isfinite_impl,
        # int64
        "l->?": npyfuncs.np_int_isfinite_impl,
        "L->?": npyfuncs.np_int_isfinite_impl,
        # intp
        "q->?": npyfuncs.np_int_isfinite_impl,
        "Q->?": npyfuncs.np_int_isfinite_impl,
        # boolean
        "?->?": npyfuncs.np_int_isfinite_impl,
        # datetime & timedelta
        "M->?": npyfuncs.np_datetime_isfinite_impl,
        "m->?": npyfuncs.np_datetime_isfinite_impl,
    }

    ufunc_db[np.signbit] = {
        "f->?": npyfuncs.np_real_signbit_impl,
        "d->?": npyfuncs.np_real_signbit_impl,
    }

    ufunc_db[np.copysign] = {
        "ff->f": npyfuncs.np_real_copysign_impl,
        "dd->d": npyfuncs.np_real_copysign_impl,
    }

    ufunc_db[np.nextafter] = {
        "ff->f": npyfuncs.np_real_nextafter_impl,
        "dd->d": npyfuncs.np_real_nextafter_impl,
    }

    ufunc_db[np.spacing] = {
        "f->f": npyfuncs.np_real_spacing_impl,
        "d->d": npyfuncs.np_real_spacing_impl,
    }

    ufunc_db[np.ldexp] = {
        "fi->f": npyfuncs.np_real_ldexp_impl,
        "fl->f": npyfuncs.np_real_ldexp_impl,
        "di->d": npyfuncs.np_real_ldexp_impl,
        "dl->d": npyfuncs.np_real_ldexp_impl,
    }
    if numpy_version >= (2, 0) and IS_WIN32:
        ufunc_db[np.ldexp]["fq->f"] = ufunc_db[np.ldexp].pop("fl->f")
        ufunc_db[np.ldexp]["dq->d"] = ufunc_db[np.ldexp].pop("dl->d")

    # bit twiddling functions
    ufunc_db[np.bitwise_and] = {
        "??->?": numbers.int_and_impl,
        "bb->b": numbers.int_and_impl,
        "BB->B": numbers.int_and_impl,
        "hh->h": numbers.int_and_impl,
        "HH->H": numbers.int_and_impl,
        "ii->i": numbers.int_and_impl,
        "II->I": numbers.int_and_impl,
        "ll->l": numbers.int_and_impl,
        "LL->L": numbers.int_and_impl,
        "qq->q": numbers.int_and_impl,
        "QQ->Q": numbers.int_and_impl,
    }

    ufunc_db[np.bitwise_or] = {
        "??->?": numbers.int_or_impl,
        "bb->b": numbers.int_or_impl,
        "BB->B": numbers.int_or_impl,
        "hh->h": numbers.int_or_impl,
        "HH->H": numbers.int_or_impl,
        "ii->i": numbers.int_or_impl,
        "II->I": numbers.int_or_impl,
        "ll->l": numbers.int_or_impl,
        "LL->L": numbers.int_or_impl,
        "qq->q": numbers.int_or_impl,
        "QQ->Q": numbers.int_or_impl,
    }

    ufunc_db[np.bitwise_xor] = {
        "??->?": numbers.int_xor_impl,
        "bb->b": numbers.int_xor_impl,
        "BB->B": numbers.int_xor_impl,
        "hh->h": numbers.int_xor_impl,
        "HH->H": numbers.int_xor_impl,
        "ii->i": numbers.int_xor_impl,
        "II->I": numbers.int_xor_impl,
        "ll->l": numbers.int_xor_impl,
        "LL->L": numbers.int_xor_impl,
        "qq->q": numbers.int_xor_impl,
        "QQ->Q": numbers.int_xor_impl,
    }

    ufunc_db[np.invert] = {  # aka np.bitwise_not
        "?->?": numbers.int_invert_impl,
        "b->b": numbers.int_invert_impl,
        "B->B": numbers.int_invert_impl,
        "h->h": numbers.int_invert_impl,
        "H->H": numbers.int_invert_impl,
        "i->i": numbers.int_invert_impl,
        "I->I": numbers.int_invert_impl,
        "l->l": numbers.int_invert_impl,
        "L->L": numbers.int_invert_impl,
        "q->q": numbers.int_invert_impl,
        "Q->Q": numbers.int_invert_impl,
    }

    ufunc_db[np.left_shift] = {
        "bb->b": numbers.int_shl_impl,
        "BB->B": numbers.int_shl_impl,
        "hh->h": numbers.int_shl_impl,
        "HH->H": numbers.int_shl_impl,
        "ii->i": numbers.int_shl_impl,
        "II->I": numbers.int_shl_impl,
        "ll->l": numbers.int_shl_impl,
        "LL->L": numbers.int_shl_impl,
        "qq->q": numbers.int_shl_impl,
        "QQ->Q": numbers.int_shl_impl,
    }

    ufunc_db[np.right_shift] = {
        "bb->b": numbers.int_shr_impl,
        "BB->B": numbers.int_shr_impl,
        "hh->h": numbers.int_shr_impl,
        "HH->H": numbers.int_shr_impl,
        "ii->i": numbers.int_shr_impl,
        "II->I": numbers.int_shr_impl,
        "ll->l": numbers.int_shr_impl,
        "LL->L": numbers.int_shr_impl,
        "qq->q": numbers.int_shr_impl,
        "QQ->Q": numbers.int_shr_impl,
    }

    # Inject datetime64 support
    from numba.cuda.np import npdatetime

    ufunc_db[np.negative].update(
        {
            "m->m": npdatetime.timedelta_neg_impl,
        }
    )
    ufunc_db[np.positive].update(
        {
            "m->m": npdatetime.timedelta_pos_impl,
        }
    )
    ufunc_db[np.absolute].update(
        {
            "m->m": npdatetime.timedelta_abs_impl,
        }
    )
    ufunc_db[np.sign].update(
        {
            "m->m": npdatetime.timedelta_sign_impl,
        }
    )
    ufunc_db[np.add].update(
        {
            "mm->m": npdatetime.timedelta_add_impl,
            "Mm->M": npdatetime.datetime_plus_timedelta,
            "mM->M": npdatetime.timedelta_plus_datetime,
        }
    )
    ufunc_db[np.subtract].update(
        {
            "mm->m": npdatetime.timedelta_sub_impl,
            "Mm->M": npdatetime.datetime_minus_timedelta,
            "MM->m": npdatetime.datetime_minus_datetime,
        }
    )
    ufunc_db[np.multiply].update(
        {
            "mq->m": npdatetime.timedelta_times_number,
            "md->m": npdatetime.timedelta_times_number,
            "qm->m": npdatetime.number_times_timedelta,
            "dm->m": npdatetime.number_times_timedelta,
        }
    )
    if np.divide != np.true_divide:
        ufunc_db[np.divide].update(
            {
                "mq->m": npdatetime.timedelta_over_number,
                "md->m": npdatetime.timedelta_over_number,
                "mm->d": npdatetime.timedelta_over_timedelta,
            }
        )
    ufunc_db[np.true_divide].update(
        {
            "mq->m": npdatetime.timedelta_over_number,
            "md->m": npdatetime.timedelta_over_number,
            "mm->d": npdatetime.timedelta_over_timedelta,
        }
    )
    ufunc_db[np.floor_divide].update(
        {
            "mq->m": npdatetime.timedelta_over_number,
            "md->m": npdatetime.timedelta_over_number,
        }
    )

    ufunc_db[np.floor_divide].update(
        {
            "mm->q": npdatetime.timedelta_floor_div_timedelta,
        }
    )

    ufunc_db[np.equal].update(
        {
            "MM->?": npdatetime.datetime_eq_datetime_impl,
            "mm->?": npdatetime.timedelta_eq_timedelta_impl,
        }
    )
    ufunc_db[np.not_equal].update(
        {
            "MM->?": npdatetime.datetime_ne_datetime_impl,
            "mm->?": npdatetime.timedelta_ne_timedelta_impl,
        }
    )
    ufunc_db[np.less].update(
        {
            "MM->?": npdatetime.datetime_lt_datetime_impl,
            "mm->?": npdatetime.timedelta_lt_timedelta_impl,
        }
    )
    ufunc_db[np.less_equal].update(
        {
            "MM->?": npdatetime.datetime_le_datetime_impl,
            "mm->?": npdatetime.timedelta_le_timedelta_impl,
        }
    )
    ufunc_db[np.greater].update(
        {
            "MM->?": npdatetime.datetime_gt_datetime_impl,
            "mm->?": npdatetime.timedelta_gt_timedelta_impl,
        }
    )
    ufunc_db[np.greater_equal].update(
        {
            "MM->?": npdatetime.datetime_ge_datetime_impl,
            "mm->?": npdatetime.timedelta_ge_timedelta_impl,
        }
    )
    ufunc_db[np.maximum].update(
        {
            "MM->M": npdatetime.datetime_maximum_impl,
            "mm->m": npdatetime.timedelta_maximum_impl,
        }
    )
    ufunc_db[np.minimum].update(
        {
            "MM->M": npdatetime.datetime_minimum_impl,
            "mm->m": npdatetime.timedelta_minimum_impl,
        }
    )
    # there is no difference for datetime/timedelta in maximum/fmax
    # and minimum/fmin
    ufunc_db[np.fmax].update(
        {
            "MM->M": npdatetime.datetime_fmax_impl,
            "mm->m": npdatetime.timedelta_fmax_impl,
        }
    )
    ufunc_db[np.fmin].update(
        {
            "MM->M": npdatetime.datetime_fmin_impl,
            "mm->m": npdatetime.timedelta_fmin_impl,
        }
    )

    ufunc_db[np.remainder].update(
        {
            "mm->m": npdatetime.timedelta_mod_timedelta,
        }
    )



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause

"""
Hints to wrap Kernel arguments to indicate how to manage host-device
memory transfers before & after the kernel call.
"""

import functools


from numba.cuda.typing.typeof import typeof, Purpose


class ArgHint:
    def __init__(self, value):
        self.value = value

    def to_device(self, retr, stream=0):
        """
        :param stream: a stream to use when copying data
        :param retr:
            a list of clean-up work to do after the kernel's been run.
            Append 0-arg lambdas to it!
        :return: a value (usually an `DeviceNDArray`) to be passed to
            the kernel
        """

    @functools.cached_property
    def _numba_type_(self):
        return typeof(self.value, Purpose.argument)


class In(ArgHint):
    def to_device(self, retr, stream=0):
        from .cudadrv.devicearray import _to_strided_memory_view

        devary, _ = _to_strided_memory_view(self.value, stream=stream)
        # A dummy writeback functor to keep devary alive until the kernel
        # is called.
        retr.append(lambda: devary)
        return devary


class Out(ArgHint):
    copy_input = False

    def to_device(self, retr, stream=0):
        from .cudadrv.devicearray import _to_strided_memory_view
        from .cudadrv.devicearray import _make_strided_memory_view
        from .cudadrv.driver import driver
        from .cudadrv import devices

        devary, conv = _to_strided_memory_view(
            value := self.value, copy=self.__class__.copy_input, stream=stream
        )
        if conv:
            stream_ptr = getattr(stream, "handle", stream)

            def copy_to_host(devary=devary, value=value, stream_ptr=stream_ptr):
                hostary = _make_strided_memory_view(
                    value, stream_ptr=stream_ptr
                )
                nbytes = devary.size * devary.dtype.itemsize
                hostptr = hostary.ptr
                devptr = devary.ptr
                if int(stream_ptr):
                    driver.cuMemcpyDtoHAsync(
                        hostptr, devptr, nbytes, stream_ptr
                    )
                else:
                    driver.cuMemcpyDtoH(hostptr, devptr, nbytes)
                    ctx = devices.get_context()
                    stream = ctx.get_default_stream()
                    stream.synchronize()
                return hostary

            retr.append(copy_to_host)
        return devary


class InOut(Out):
    copy_input = True


def wrap_arg(value, default=InOut):
    return value if isinstance(value, ArgHint) else default(value)


__all__ = [
    "In",
    "Out",
    "InOut",
    "ArgHint",
    "wrap_arg",
]



# SPDX-FileCopyrightText: Copyright (c) 2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause

from numba_cuda._version import __version__

__all__ = ["__version__"]
